
#include "x.h"
#include "xt.h"
#include "xm.h"
#include "xpex.h"
#include "off.h"
#include "unix.h"
#include "book_utils.h"
#include "defs.h"
#include "externs.h"
#include "color.h"
#include "transformation.h"
#include "dialog.h"

typedef struct _Vertex {
    float x, y, z;
} 
Vertex;

typedef Vertex Vector;

typedef struct _Edge {
    int e;
} 
Edge;

typedef struct _Facet {
    int count;
    Edge *edges;
} 
Facet;

extern XtActionProc CanvasButtonPress(), CanvasButtonRelease(),
	CanvasB1Motion(), CanvasB2Motion(), CanvasB3Motion(),
	CanvasMessage(), CanvasExpose(), CanvasConfigure(), CanvasResize(),
	CanvasUp(), CanvasDown(), CanvasLeft(), CanvasRight(),
	CanvasReset(),
	CanvasWireframe(), CanvasCull(), CanvasEdge(), CanvasShadow(),
	CanvasColor(), CanvasMaterial(), CanvasMaterialPredef(), CanvasLight(),
	CanvasFlyby(), CanvasAnimateCr(), CanvasAnimateEx(),
    CanvasMorph(), CanvasSpin(),
	CanvasTaper(), CanvasTwist(), CanvasBend(),
	CanvasStepBack(), CanvasPlay(), CanvasStepForw(), CanvasStop(),
    CanvasRecord(),
    CanvasSavePPM(), CanvasSaveMPG(),
	CanvasHardcopy(),
    CanvasMaterialPredefQuit(), CanvasVideoQuit(), CanvasQuit();

extern Widget toplevel;

int lastx, lasty;
float theta, phi;
int motion;
float curquat[4];
int ppm = 0;
int mpg = 0;

#define TWO_PI 6.283185308
#define DEGRAD( _a )  ( (double)( _a ) * 3.1415926 / 180.0 )
double  rotA = DEGRAD(-60.0), rotB = DEGRAD(0.0);

#define NORMALIZE( _v ) \
{ \
    double    mag; \
\
    mag = 1.0 / sqrt((_v).x * (_v).x \
          + (_v).y * (_v).y + (_v).z * (_v).z); \
    (_v).x *= mag, (_v).y *= mag, (_v).z *= mag; \
}

XtInputId colorInput, matInput;

extern XtAppContext      app_context;
extern Display           *display;
extern Window            window;
extern Pixmap            pixmap;
extern XStandardColormap cmap_info;
extern GC                gc;
       Cursor            cursor;
extern PEXRenderer       renderer;
extern PEXStructure      top_struct, object, object_shadow,
                         empty, plane, axis;
extern XWindowAttributes xwattrs;
extern int               using_drawable;
extern Drawable          drawable;
extern Multibuffer       buffers[2];

extern char *FileName;

extern Boolean colorTypeIndexed, colorTypeRGB,
colorTypeHSV, colorTypeHLS, colorTypeCIE;

typedef struct _movie *Movie;

struct _movie {
#ifdef PIX_FRAME
    Pixmap frame;
#else
    XImage *frame;
#endif
    Movie next;
    Movie prev;
};

Movie frame, head, copy_head;


void
add_frame( drawable )
Drawable drawable;
{
    extern XVisualInfo vis_info;

#ifdef PIX_FRAME
    frame->frame = XCreatePixmap( display, drawable,
        xwattrs.width, xwattrs.height, vis_info.depth );

    if( frame->frame == (Pixmap)NULL ) {
        fprintf( stderr, "Cant allocate a pixmap.\n" );
        exit(1);
    }
    XCopyArea( display, drawable, frame->frame, gc, 0, 0,
        xwattrs.width, xwattrs.height, 0, 0 );
#else
    frame->frame = XCreateImage( display, vis_info.visual, vis_info.depth,
        ZPixmap, 0, (char *)0, xwattrs.width, xwattrs.height,
        ((vis_info.depth>8)?32:8), 0 );
    if( frame->frame == (XImage *)NULL ) {
        fprintf( stderr, "Cant allocate an image.\n" );
        exit(1);
    }
#if( 0 ) /* this is not needed. XGetImage allocs it for you */
    frame->frame->data = malloc( frame->frame->bytes_per_line * 
                                 frame->frame->height );
    if( frame->frame->data == 0 ) {
        fprintf( stderr, "Cant allocate image data\n" );
        exit(1);
    }
#endif
    frame->frame = XGetImage( display, drawable, 0, 0,
        xwattrs.width, xwattrs.height, -1, ZPixmap );
#endif

    if((frame->next = (Movie) malloc( sizeof(struct _movie) )) == NULL) {
        fprintf( stderr, "add_frame: ran out of memory!\n" );
        exit(0);
    }
    frame->next->prev = frame;
    frame = frame->next;
    frame->frame = 0;
    frame->next = NULL;
}


void
redraw()
{
    extern int recording;

    PEXRenderNetwork( display, drawable, renderer, top_struct );
    XFlush( display );

    if( using_drawable == BUFFER )
        /* Display buffer */
        XmbufDisplayBuffers( display, 1, &drawable, 0, 0 );
    else if( using_drawable == PIXMAP )
        /* Copy the entire pixmap to the window. */
        XCopyArea( display, drawable, window, gc, 0, 0,
            (unsigned int)xwattrs.width, (unsigned int)xwattrs.height,
            0, 0 );
    else if( using_drawable == WINDOW ) ;

    if( recording ) add_frame( drawable );

    if( using_drawable == BUFFER )
        /* Swap buffer */
        drawable = (drawable == buffers[0] ? buffers[1] : buffers[0]);
}


XtActionProc
CanvasMessage( w, event )
Widget    w;
XEvent   *event;
{
    extern Atom xa_WM_PROTOCOLS;
    extern Atom xa_WM_DELETE_WINDOW;

    if( event->xclient.message_type == xa_WM_PROTOCOLS )
        if( event->xclient.data.l[0] == xa_WM_DELETE_WINDOW )
            CanvasQuit( w, event );
}


XtActionProc
CanvasExpose( w, event )
Widget    w;
XEvent   *event;
{
    /* Redraw when last Expose event received. */
    if ( event->xexpose.count == 0 )
        redraw();
}


XtActionProc
CanvasConfigure( w, event )
Widget    w;
XEvent   *event;
{
}


XtActionProc
CanvasResize( w, event )
Widget    w;
XEvent   *event;
{
    if( using_drawable == PIXMAP ) {
        /* Compare old size to the new size */
        if( event->xconfigure.width != xwattrs.width ||
            event->xconfigure.height != xwattrs.height ) {

            /* The size changed */
            XFreePixmap( display, pixmap );

            /* Create a new pixmap at the new window size */
            XGetWindowAttributes( display, window, &xwattrs );
            pixmap = XCreatePixmap( display, window,
                (unsigned int)xwattrs.width, (unsigned int)xwattrs.height,
                (unsigned int)xwattrs.depth );
        }
    }
}


XtActionProc
CanvasButtonPress( w, event )
Widget    w;
XEvent   *event;
{
    PEXRendererAttributes rattrs;
    extern PEXRenderer	renderer;
    extern int no_axis;
    extern PEXVector	view_plane_normal;

    lastx  = ((XButtonEvent *)event)->x;
    lasty  = xwattrs.height - ((XButtonEvent *)event)->y;
    lasty  = ((XButtonEvent *)event)->y;

    /*
     * atan2(sin(theta), cos(theta)) gives theta.
     * since we have sin(phi) = dy and dx = cos(phi)*sin(theta)
     * we want atan2(dy, dx/sin(theta)) BUT sin(theta) will be zero
     * so we use dy*sin(theta) BUT when sin(theta) and dx are negative
     * it changes the quadrant. so...
     */
    theta = atan2( view_plane_normal.x,  view_plane_normal.z );
    if( sin(theta) >= 0 )
        phi = atan2(  view_plane_normal.y * sin(theta),  view_plane_normal.x);
    else
        phi = atan2( -view_plane_normal.y * sin(theta), -view_plane_normal.x);

    /* UnSet the renderer's clearI flag */
    if( ((XButtonEvent *)event)->button == 1 ) {
        cursor = XCreateFontCursor( display, XC_left_ptr );
        XDefineCursor( display, window, cursor );
        XFlush( display );

        if( !no_axis ) {
        rattrs.clear_image = 0;
        PEXChangeRenderer( display, renderer, PEXRAClearImage, &rattrs );
        if( using_drawable == BUFFER )
            /* Swap buffer */
            drawable = (drawable == buffers[0] ? buffers[1] : buffers[0]);
        PEXRenderNetwork( display, drawable, renderer, axis );
        }
    }
    else if( ((XButtonEvent *)event)->button == 2 ) {
        cursor = XCreateFontCursor( display, XC_fleur );
        XDefineCursor( display, window, cursor );
    }
    else if( ((XButtonEvent *)event)->button == 3 ) {
        cursor = XCreateFontCursor( display, XC_sizing );
        XDefineCursor( display, window, cursor );
    }
    XFlush( display );
/*  trackball(curquat, 0.0, 0.0, 0.0, 0.0); */
}


XtActionProc
CanvasButtonRelease( w, event )
Widget    w;
XEvent   *event;
{
    PEXRendererAttributes rattrs;
    extern PEXRenderer renderer;
    extern PEXMatrix local_matrix, shad_local_matrix;
    extern int no_axis;

    XUndefineCursor( display, window );
    XFreeCursor( display, cursor );

    /* Update the display */
    /* Set the renderer's clearI flag */
    if( !no_axis ) {
    if( ((XButtonEvent *)event)->button == 1 ) {
        rattrs.clear_image = 1;
        PEXChangeRenderer( display, renderer, PEXRAClearImage, &rattrs );
        if( using_drawable == BUFFER )
            drawable = (drawable == buffers[0] ? buffers[1] : buffers[0]);
    }
    redraw();
    }
}


XtActionProc
CanvasB1Motion( w, event )
Widget    w;
XEvent   *event;
{
    extern PEXMatrix local_matrix, shad_local_matrix;
    extern PEXRendererAttributes rattrs;
    PEXMatrix	rot_matrix;
    float		newquat[4];
    extern int no_axis;

    PEXVector	vpn, vup;
    float		cx, cy, cz;
    int t, c;

    extern PEXCoord     view_ref_pt, view_ref_pt_a;
    extern PEXVector    view_up_vec, view_up_vec_a;
    extern PEXVector    view_plane_normal, view_plane_normal_a;
    extern PEXViewEntry view, view_a;

    theta += 0.005 * (lastx - ((XMotionEvent *)event)->x);
    phi   += 0.005 * (((XMotionEvent *)event)->y - lasty);

    view_plane_normal_a.x = cos( phi ) * sin( theta );
    view_plane_normal_a.y = sin( phi );
    view_plane_normal_a.z = cos( phi ) * cos( theta );

    vup.x = view_up_vec_a.x;
    vup.y = view_up_vec_a.y;
    vup.z = view_up_vec_a.z;

    t = 6; c = 0;
    while (t--) {
        cx = view_plane_normal_a.y * view_up_vec_a.z -
             view_plane_normal_a.z * view_up_vec_a.y;
        cy = view_plane_normal_a.z * view_up_vec_a.x -
             view_plane_normal_a.x * view_up_vec_a.z;
        cz = view_plane_normal_a.x * view_up_vec_a.y -
             view_plane_normal_a.y * view_up_vec_a.x;

        if( (cx*cx + cy*cy + cz*cz) < 0.1 ) {
            if( (view_up_vec_a.y < 0.8) && (view_up_vec_a.y > -0.8) )
                view_up_vec_a.y += 0.03;
            else
                view_up_vec_a.x += 0.05;
            NORMALIZE(view_up_vec_a);
            c = 1;
        }
        else
            t = 0;
    }
    if( !c ) {
        view_up_vec_a.x = vup.x;
        view_up_vec_a.y = vup.y;
        view_up_vec_a.z = vup.z;
    }

    view_plane_normal.x = view_plane_normal_a.x;
    view_plane_normal.y = view_plane_normal_a.y;
    view_plane_normal.z = view_plane_normal_a.z;
    view_up_vec.x = view_up_vec_a.x;
    view_up_vec.y = view_up_vec_a.y;
    view_up_vec.z = view_up_vec_a.z;

#if( 0 )	/*  Mathod 1 */
    rotA = (TWO_PI * 
            ((xwattrs.height - ((XMotionEvent *)event)->y) - lasty)) / 
           xwattrs.height;
    rotB = (TWO_PI * (((XMotionEvent *)event)->x - lastx)) / xwattrs.width;
#endif		/*  Method 1 */
#if( 0 )	/*  Method 2 */
    rotate_object( (double)lastx,
                   (double)lasty,
                   (double)((XMotionEvent *)event)->x, 
                   (double)(xwattrs.height - ((XMotionEvent *)event)->y),
                   rot_matrix );
#endif		/*  Method 2 */
#if( 0 )	/*  Method 3 */
    trackball( newquat,
        2.0*lastx/xwattrs.width - 1.0,
        2.0*(xwattrs.height-lasty)/xwattrs.height - 1.0,
        2.0*((XMotionEvent *)event)->x/xwattrs.width - 1.0,
        2.0*(xwattrs.height-((XMotionEvent *)event)->y)/xwattrs.height - 1.0 );
    add_quats( newquat, curquat, curquat );
    build_rotmatrix(rot_matrix, curquat);
#endif		/*  Method 3 */
    lastx = ((XMotionEvent *)event)->x;
    lasty = xwattrs.height - ((XMotionEvent *)event)->y;
    lasty = ((XMotionEvent *)event)->y;
#if( 0 )	/*  Method 1 */
    PEXRotate( PEXXAxis, rotA, rot_matrix );
    PEXMatrixMult( rot_matrix, local_matrix, local_matrix );
    PEXMatrixMult( rot_matrix, shad_local_matrix, shad_local_matrix );

    PEXRotate( PEXYAxis, rotB, rot_matrix );
    PEXMatrixMult( rot_matrix, local_matrix, local_matrix );
    PEXMatrixMult( rot_matrix, shad_local_matrix, shad_local_matrix );
#endif		/*  Mathod 1 */
#if( 0 )	/*  Method 2 3 */
    PEXMatrixMult( rot_matrix, local_matrix, local_matrix );
    PEXMatrixMult( rot_matrix, shad_local_matrix, shad_local_matrix );
#endif		/*  Mathod 2 3 */

    PEXViewOrientationMatrix( &view_ref_pt, &view_plane_normal,
        &view_up_vec, view.orientation );
    PEXSetTableEntries( display, rattrs.view_table, 1, 1, PEXLUTView, &view );

    PEXViewOrientationMatrix( &view_ref_pt_a, &view_plane_normal_a,
        &view_up_vec_a, view_a.orientation );
    PEXSetTableEntries( display, rattrs.view_table, 3, 1, PEXLUTView, &view_a );
    if( !no_axis ) {
    PEXRenderNetwork( display, drawable, renderer, axis );
    XFlush( display );
    PEXSetTableEntries( display, rattrs.view_table, 4, 1, PEXLUTView, &view_a );
    }
    else
    redraw();
}


XtActionProc
CanvasB2Motion( w, event )
Widget    w;
XEvent   *event;
{
    extern PEXRendererAttributes rattrs;
    extern PEXCoord2D       view_window[2];
    extern PEXNPCSubVolume  viewport;
    extern double           view_plane, front_plane, back_plane;
    extern PEXCoord         proj_ref_point;
    extern int              perspective;
    extern PEXViewEntry     view;

    int dx, dy;
    float zoom_factor = 0.5 * 0.01;

    dx = lastx - ((XMotionEvent *)event)->x;
    dy = ((XMotionEvent *)event)->y - lasty;

    view_window[0].x += (float)dx * zoom_factor;
    view_window[1].x += (float)dx * zoom_factor;
    view_window[0].y += (float)dy * zoom_factor;
    view_window[1].y += (float)dy * zoom_factor;

    PEXViewMappingMatrix( view_window, &viewport, perspective,
        &proj_ref_point, view_plane, back_plane, front_plane, view.mapping );
    PEXSetTableEntries( display, rattrs.view_table, 1, 1, PEXLUTView, &view );

    /* Update the display */
    redraw();

    lastx = ((XButtonEvent *)event)->x;
    lasty = ((XButtonEvent *)event)->y;
}


XtActionProc
CanvasB3Motion( w, event )
Widget    w;
XEvent   *event;
{
    extern PEXRendererAttributes rattrs;
    extern PEXCoord2D       view_window[2];
    extern PEXNPCSubVolume  viewport;
    extern double           view_plane, front_plane, back_plane;
    extern PEXCoord         proj_ref_point;
    extern int              perspective;
    extern PEXViewEntry     view;

    float old_width, old_height;
    float zoom_factor = 0.5 * 0.1;

    old_width = view_window[1].x - view_window[0].x;
    old_height = view_window[1].y - view_window[0].y;

    if( ((XButtonEvent *)event)->y < lasty ) {
        /* Zoom out */
        view_window[0].x -= zoom_factor * old_width;
        view_window[1].x += zoom_factor * old_width;
        view_window[0].y -= zoom_factor * old_height;
        view_window[1].y += zoom_factor * old_height;
    }
    else {
        /* Zoom in */
        view_window[0].x += zoom_factor * old_width;
        view_window[1].x -= zoom_factor * old_width;
        view_window[0].y += zoom_factor * old_height;
        view_window[1].y -= zoom_factor * old_height;
    }

    PEXViewMappingMatrix( view_window, &viewport, perspective,
        &proj_ref_point, view_plane, back_plane, front_plane, view.mapping );
    PEXSetTableEntries( display, rattrs.view_table, 1, 1, PEXLUTView, &view );

    /* Update the display */
    redraw();

    lastx = ((XButtonEvent *)event)->x;
    lasty = ((XButtonEvent *)event)->y;
}


XtActionProc
CanvasReset()
{
    extern PEXMatrix init_matrix, shad_local_matrix;

    initialize_view();

    PEXSetEditingMode( display, object, PEXStructureReplace );
    PEXSetElementPtr( display, object, PEXBeginning, 0 );
    PEXSetElementPtrAtLabel( display, object, ROTATION_LABEL, 1 );
    PEXSetLocalTransform( display, object, PEXOCStore,
        PEXReplace, init_matrix );

    {
        PEXCoord4D  light;
        PEXMatrix   new_matrix;
        light.x = 10.0;
        light.y = 100.0;
        light.z = 7.0;
        light.w = PEXLightInfinite;
        /* This should have been -3.49, but it doesnt look good */
        PEXUmbraMatrix( PEXYPlane, -2.49, light, shad_local_matrix );
        PEXSetEditingMode( display, object_shadow, PEXStructureReplace );
        PEXSetElementPtr( display, object_shadow, PEXBeginning, 0 );
        PEXSetElementPtrAtLabel( display, object_shadow,
            SHADOW_ROTATION_LABEL, 1 );
        PEXSetLocalTransform( display, object_shadow, PEXOCStore,
            PEXReplace, shad_local_matrix );
    }

    redraw();
}


XtActionProc
CanvasCull()
{
    extern Widget w_cull;
    extern int culling;
    culling = 1 - culling;

    PEXSetEditingMode( display, top_struct, PEXStructureReplace );
    PEXSetElementPtr( display, top_struct, PEXBeginning, 0 );
    PEXSetElementPtrAtLabel( display, top_struct, CULL_LABEL, 1 );
    if( culling ) {
        XtVaSetValues( w_cull,
            XmNlabelString, XmStringCreateSimple( "Cull On" ), NULL );
        PEXSetFacetCullingMode( display, top_struct, PEXOCStore, PEXBackFaces );
    }
    else {
        XtVaSetValues( w_cull,
            XmNlabelString, XmStringCreateSimple( "Cull Off" ), NULL );
        PEXSetFacetCullingMode( display, top_struct, PEXOCStore, PEXNone );
    }
    redraw();
}


XtActionProc
CanvasEdge()
{
    extern Widget w_edge;
    extern int edges;
    edges = 1 - edges;

    PEXSetEditingMode( display, top_struct, PEXStructureReplace );
    PEXSetElementPtr( display, top_struct, PEXBeginning, 0 );
    PEXSetElementPtrAtLabel( display, top_struct, WIREFRAME_LABEL, 1 );
    if( edges ) {
        XtVaSetValues( w_edge,
            XmNlabelString, XmStringCreateSimple( "Edges On" ), NULL );
        PEXSetSurfaceEdgeFlag( display, top_struct, PEXOCStore, PEXOn );
    }
    else {
        XtVaSetValues( w_edge,
            XmNlabelString, XmStringCreateSimple( "Edges Off" ), NULL );
        PEXSetSurfaceEdgeFlag( display, top_struct, PEXOCStore, PEXOff );
    }
    redraw();
}


XtActionProc
CanvasWireframe(Widget w, XEvent *event)
{
    extern Widget w_wireframe;
    extern int wireframe, edges;
    wireframe = 1 - wireframe;

    PEXSetEditingMode( display, top_struct, PEXStructureReplace );
    PEXSetElementPtr( display, top_struct, PEXBeginning, 0 );

    /* the order of these calls is important */
    /* depends on the order the labels were put into the structure */
    if( wireframe ) {
        XtVaSetValues( w_wireframe,
            XmNlabelString, XmStringCreateSimple( "Wire" ), NULL );

        PEXSetElementPtrAtLabel( display, top_struct, DEPTH_CUE_LABEL, 1 );
        PEXSetDepthCueIndex( display, top_struct, PEXOCStore, 1 );

        PEXSetElementPtrAtLabel( display, top_struct, WIREFRAME_LABEL, 1 );
        PEXSetSurfaceEdgeFlag( display, top_struct, PEXOCStore, PEXOn );

        PEXSetElementPtrAtLabel( display, top_struct, SHADING_METHOD_LABEL, 1 );
        PEXSetSurfaceInterpMethod( display, top_struct, PEXOCStore,
            PEXSurfaceInterpNone );

        PEXSetElementPtrAtLabel( display, top_struct, INTERIOR_LABEL, 1 );
        PEXSetInteriorStyle( display, top_struct, PEXOCStore,
            PEXInteriorStyleHollow );
        PEXSetBFInteriorStyle( display, top_struct, PEXOCStore,
            PEXInteriorStyleHollow );

        PEXSetElementPtrAtLabel( display, top_struct, REFL_EQN_LABEL, 1 );
        PEXSetReflectionModel( display, top_struct, PEXOCStore,
            PEXReflectionNone );
    }
    else {
        XtVaSetValues( w_wireframe,
            XmNlabelString, XmStringCreateSimple( "Solid" ), NULL );

        PEXSetElementPtrAtLabel( display, top_struct, DEPTH_CUE_LABEL, 1 );
        PEXSetDepthCueIndex( display, top_struct, PEXOCStore, 0 );

        PEXSetElementPtrAtLabel( display, top_struct, WIREFRAME_LABEL, 1 );
        if( edges )
        PEXSetSurfaceEdgeFlag( display, top_struct, PEXOCStore, PEXOn );
        else
        PEXSetSurfaceEdgeFlag( display, top_struct, PEXOCStore, PEXOff );

        PEXSetElementPtrAtLabel( display, top_struct, SHADING_METHOD_LABEL, 1 );
        PEXSetSurfaceInterpMethod( display, top_struct, PEXOCStore,
            PEXSurfaceInterpColor );

        PEXSetElementPtrAtLabel( display, top_struct, INTERIOR_LABEL, 1 );
        PEXSetInteriorStyle( display, top_struct, PEXOCStore,
            PEXInteriorStyleSolid );
        PEXSetBFInteriorStyle( display, top_struct, PEXOCStore,
            PEXInteriorStyleSolid );

        PEXSetElementPtrAtLabel( display, top_struct, REFL_EQN_LABEL, 1 );
        PEXSetReflectionModel( display, top_struct, PEXOCStore,
            PEXReflectionSpecular );
    }
    redraw();
}


XtActionProc
CanvasShadow()
{
    extern int draw_shadow;
    draw_shadow = 1 - draw_shadow;

    PEXSetEditingMode( display, top_struct, PEXStructureReplace );
    PEXSetElementPtr( display, top_struct, PEXBeginning, 0 );
    PEXSetElementPtrAtLabel( display, top_struct, SHADOW_LABEL, 1 );
    if( draw_shadow )
        PEXExecuteStructure( display, top_struct, PEXOCStore, object_shadow );
    else
        PEXNoop( display, top_struct, PEXOCStore );

    redraw();
}


XtActionProc
CanvasColor( Widget w, XEvent *event )
{
    extern char *getenv();
    extern PEXColor colorMaterial;
    extern void get_color_reply();

    if( getenv( "COLOR_EDITOR" ) ) {
        char *program, *args;
        extern int smth_changed;

        smth_changed = 1;

        colorInput = XtAppAddInput( app_context, fileno(stdin),
            (XtPointer)XtInputReadMask, get_color_reply, NULL );

        program = (char *) malloc( strlen(getenv( "COLOR_EDITOR" )) + 25 );
        strcpy( program, getenv( "COLOR_EDITOR" ) );
        args = (char *) malloc(25);
        sprintf( args, " RGB %.2f %.2f %.2f", colorMaterial.rgb.red,
            colorMaterial.rgb.green, colorMaterial.rgb.blue );
        strcat( program, args );
        xs_talkto( program );
    }
    else {
        fprintf( stderr, "Environment variable COLOR_EDITOR, undefined!!!\n\n");
        XtDestroyApplicationContext( app_context );
        exit( 0 );
    }
}


XtActionProc
CanvasMaterial(Widget w, XEvent *event)
{
    extern char *getenv();
    extern PEXColor colorMaterial;
    extern PEXReflectionAttributes refl_props;
    extern void get_material_reply();

    if( getenv( "MATERIAL_EDITOR" ) ) {
        char *program, *args;
        extern int smth_changed;

        smth_changed = 1;

        matInput = XtAppAddInput( app_context, fileno(stdin),
            (XtPointer)XtInputReadMask, get_material_reply, NULL );

        program = (char *) malloc( strlen(getenv( "MATERIAL_EDITOR" )) + 80 );
        strcpy( program, getenv( "MATERIAL_EDITOR" ) );
        args = (char *) malloc(80);
        sprintf( args, " %.2f %.2f %.2f %.2f %.2f RGB %.2f %.2f %.2f RGB %.2f %.2f %.2f\n",
            refl_props.ambient, refl_props.diffuse, refl_props.transmission,
            refl_props.specular, refl_props.specular_conc,
            refl_props.specular_color.value.rgb.red,
            refl_props.specular_color.value.rgb.green,
            refl_props.specular_color.value.rgb.blue,
            colorMaterial.rgb.red, colorMaterial.rgb.green,
            colorMaterial.rgb.blue );
        strcat( program, args );
        xs_talkto( program );
    }
    else {
        fprintf( stderr,
            "Environment variable MATERIAL_EDITOR, undefined!!!\n\n");
        XtDestroyApplicationContext( app_context );
        exit( 0 );
    }
}


XtActionProc
CanvasMaterialPredef( w, client_data, call_data )
Widget  w;
caddr_t client_data, call_data;
{
    typedef struct _material_predef {
        PEXColorRGB ambientColor;
        PEXColorRGB diffuseColor;
        PEXColorRGB specularColor;
        float       shininess;
        float       transparency;
        float       shininess_conc;
    } material_predef;

    extern material_predef convert[45];
    extern int smth_changed;

    int i;
    material_predef mtl;

    PEXColor                colorMaterial;
    PEXReflectionAttributes refl_props;
    PEXLightEntry           lights[4];
    extern PEXRendererAttributes rattrs;

    smth_changed = 1;

    sscanf( client_data, "%d", &i );
    mtl = convert[i];

    PEXSetEditingMode( display, top_struct, PEXStructureReplace );
    PEXSetElementPtr( display, top_struct, PEXBeginning, 0 );
    PEXSetElementPtrAtLabel( display, top_struct, MATERIAL_COLOR_LABEL, 1 );
    SET_COLOR( mtl.ambientColor.red, mtl.ambientColor.green,
        mtl.ambientColor.blue, colorMaterial );
    PEXSetSurfaceColor( display, top_struct, PEXOCStore,
        PEXColorTypeRGB, &colorMaterial );

    lights[DIRECTIONAL].type = PEXLightWCVector;
    lights[DIRECTIONAL].direction.x = -1;
    lights[DIRECTIONAL].direction.y =  0;
    lights[DIRECTIONAL].direction.z =  0;
    NORMALIZE( lights[DIRECTIONAL].direction );
    lights[DIRECTIONAL].color.type = PEXColorTypeRGB;
    SET_COLOR( mtl.diffuseColor.red, mtl.diffuseColor.green,
        mtl.diffuseColor.blue, lights[DIRECTIONAL].color.value );
    PEXSetTableEntries( display, rattrs.light_table, 2, 1,
        PEXLUTLight, &lights[DIRECTIONAL] );

    PEXSetEditingMode( display, top_struct, PEXStructureReplace );
    PEXSetElementPtr( display, top_struct, PEXBeginning, 0 );
    PEXSetElementPtrAtLabel( display, top_struct, REFL_PROPS_LABEL, 1 );
    refl_props.ambient = 0.6;
    refl_props.diffuse = 0.6;
    refl_props.transmission  = mtl.transparency;
    refl_props.specular      = mtl.shininess;
    refl_props.specular_conc = mtl.shininess_conc;
    refl_props.specular_color.type = PEXColorTypeRGB;
    refl_props.specular_color.value.rgb.red   = mtl.specularColor.red;
    refl_props.specular_color.value.rgb.green = mtl.specularColor.green;
    refl_props.specular_color.value.rgb.blue  = mtl.specularColor.blue;
    PEXSetReflectionAttributes( display, top_struct, PEXOCStore, &refl_props );

    redraw();
}


XtActionProc
CanvasLight(Widget w, XEvent *event)
{
    extern char *getenv();
    extern void get_light_reply();
    extern int numLightsON, numLightsOFF;
    extern PEXTableIndex lightsON[4], lightsOFF[4];
    extern PEXLightEntry lights[4];
    extern int cur_shading, cur_refl;
    register int i, j;

    if( getenv( "LIGHT_EDITOR" ) ) {
        char *program, *args;
        char *targs;
        extern int smth_changed;

        smth_changed = 1;

        matInput = XtAppAddInput( app_context, fileno(stdin),
            (XtPointer)XtInputReadMask, get_light_reply, NULL );

        program = (char *) malloc( strlen(getenv( "LIGHT_EDITOR" )) + 300 );
        strcpy( program, getenv( "LIGHT_EDITOR" ) );
        args = (char *) malloc(300);
        sprintf( args, " Lgts ON %d ", numLightsON );
        for( i = 0; i < numLightsON; i++ ) {
            targs = (char *) malloc(80);
            for( j=0; j<4; j++ ) {
                if( lightsON[i] == j+1 ) {
                    switch ( j+1 ) {
                    case PEXLightAmbient:
                        sprintf( targs, "Amb RGB %.2f %.2f %.2f ",
                            lights[AMBIENT].color.value.rgb.red,
                            lights[AMBIENT].color.value.rgb.green,
                            lights[AMBIENT].color.value.rgb.blue );
                        break;
                    case PEXLightWCVector:
                        sprintf( targs, "Dir %.2f %.2f %.2f RGB %.2f %.2f %.2f ",
                            lights[DIRECTIONAL].direction.x,
                            lights[DIRECTIONAL].direction.y,
                            lights[DIRECTIONAL].direction.z,
                            lights[DIRECTIONAL].color.value.rgb.red,
                            lights[DIRECTIONAL].color.value.rgb.green,
                            lights[DIRECTIONAL].color.value.rgb.blue );
                        break;
                    case PEXLightWCPoint:
                        sprintf( targs, "Pos %.2f %.2f %.2f %.2f %.2f RGB %.2f %.2f %.2f ",
                            lights[POSITIONAL].point.x,
                            lights[POSITIONAL].point.y,
                            lights[POSITIONAL].point.z,
                            lights[POSITIONAL].attenuation1,
                            lights[POSITIONAL].attenuation2,
                            lights[POSITIONAL].color.value.rgb.red,
                            lights[POSITIONAL].color.value.rgb.green,
                            lights[POSITIONAL].color.value.rgb.blue );
                        break;
                    case PEXLightWCSpot:
                        sprintf( targs, "Spt %.2f %.2f %.2f %.2f %.2f %.2f %.2f %.2f %.2f %.2f RGB %.2f %.2f %.2f ",
                            lights[SPOT].point.x, lights[SPOT].point.y,
                            lights[SPOT].point.z, lights[SPOT].direction.x,
                            lights[SPOT].direction.y, lights[SPOT].direction.z,
                            lights[SPOT].attenuation1,
                            lights[SPOT].attenuation2,
                            lights[SPOT].concentration,
                            lights[SPOT].spread_angle,
                            lights[SPOT].color.value.rgb.red,
                            lights[SPOT].color.value.rgb.green,
                            lights[SPOT].color.value.rgb.blue );
                        break;
                    }
                }
            }
            strcat( args, targs );
        }
        targs = (char *) malloc(20);
        sprintf( targs, "Lgts OF %d ", numLightsOFF );
        strcat( args, targs );
        for( i = 0; i < numLightsOFF; i++ ) {
            targs = (char *) malloc(80);
            for( j=0; j<4; j++ ) {
                if( lightsOFF[i] == j+1 ) {
                    switch ( j+1 ) {
                    case PEXLightAmbient:
                        sprintf( targs, "Amb RGB %.2f %.2f %.2f ",
                            lights[AMBIENT].color.value.rgb.red,
                            lights[AMBIENT].color.value.rgb.green,
                            lights[AMBIENT].color.value.rgb.blue );
                        break;
                    case PEXLightWCVector:
                        sprintf( targs, "Dir %.2f %.2f %.2f RGB %.2f %.2f %.2f ",
                            lights[DIRECTIONAL].direction.x,
                            lights[DIRECTIONAL].direction.y,
                            lights[DIRECTIONAL].direction.z,
                            lights[DIRECTIONAL].color.value.rgb.red,
                            lights[DIRECTIONAL].color.value.rgb.green,
                            lights[DIRECTIONAL].color.value.rgb.blue );
                        break;
                    case PEXLightWCPoint:
                        sprintf( targs, "Pos %.2f %.2f %.2f %.2f %.2f RGB %.2f %.2f %.2f ",
                            lights[POSITIONAL].point.x,
                            lights[POSITIONAL].point.y,
                            lights[POSITIONAL].point.z,
                            lights[POSITIONAL].attenuation1,
                            lights[POSITIONAL].attenuation2,
                            lights[POSITIONAL].color.value.rgb.red,
                            lights[POSITIONAL].color.value.rgb.green,
                            lights[POSITIONAL].color.value.rgb.blue );
                        break;
                    case PEXLightWCSpot:
                        sprintf( targs, "Spt %.2f %.2f %.2f %.2f %.2f %.2f %.2f %.2f %.2f %.2f RGB %.2f %.2f %.2f ",
                            lights[SPOT].point.x, lights[SPOT].point.y,
                            lights[SPOT].point.z, lights[SPOT].direction.x,
                            lights[SPOT].direction.y, lights[SPOT].direction.z,
                            lights[SPOT].attenuation1,
                            lights[SPOT].attenuation2,
                            lights[SPOT].concentration,
                            lights[SPOT].spread_angle,
                            lights[SPOT].color.value.rgb.red,
                            lights[SPOT].color.value.rgb.green,
                            lights[SPOT].color.value.rgb.blue );
                        break;
                    }
                }
            }
            strcat( args, targs );
        }
        targs = (char *) malloc(80);
        switch ( cur_shading ) {
        case PEXSurfaceInterpNone:       
            sprintf( targs, "Shad Non " ); 
            break;
        case PEXSurfaceInterpColor:      
            sprintf( targs, "Shad Clr " ); 
            break;
        case PEXSurfaceInterpDotProduct: 
            sprintf( targs, "Shad Dot " ); 
            break;
        case PEXSurfaceInterpNormal:     
            sprintf( targs, "Shad Nor " ); 
            break;
        }
        strcat( args, targs );
        targs = (char *) malloc(80);
        switch ( cur_refl ) {
        case PEXReflectionNone:          
            sprintf( targs, "Refl Non " ); 
            break;
        case PEXReflectionAmbient:       
            sprintf( targs, "Refl Amb " ); 
            break;
        case PEXReflectionDiffuse:       
            sprintf( targs, "Refl Dif " ); 
            break;
        case PEXReflectionSpecular:      
            sprintf( targs, "Refl Spc " ); 
            break;
        }
        strcat( args, targs );
        strcat( program, args );
        xs_talkto( program );
    }
    else {
        fprintf( stderr, "Environment variable LIGHT_EDITOR, undefined!!!\n\n ");
        XtDestroyApplicationContext( app_context );
        exit( 0 );
    }
}


XtActionProc
CanvasFlyby( w, event )
Widget    w;
XEvent   *event;
{
    extern PEXVector *splineCurve();
    int n_spline_points = 0;
    PEXVector *spline_points;

    register int i, j;
    extern int n_prp_points;
    extern PEXVector *prp_points;
    extern int n_vpn_points;
    extern PEXVector *vpn_points;

    extern PEXRendererAttributes rattrs;
    extern PEXCoord         view_ref_pt;
    extern PEXVector        view_up_vec;
    extern PEXVector        view_plane_normal;
    extern PEXCoord2D       view_window[2];
    extern PEXNPCSubVolume  viewport;
    extern double           view_plane, front_plane, back_plane;
    extern PEXCoord         proj_ref_point;
    extern int              perspective;
    extern PEXViewEntry     view;

    if( n_prp_points == 0 ) {
        char *filename, *str;

        filename = (char *)malloc( strlen( FileName ) + 1 );
        strcpy( filename, FileName );
        filename[strlen(filename)-5] = '\0';

        str = (char *)malloc( 600 );
        sprintf( str, "\n\
Attribute \"flyby_prp\" is not defined in %s.aoff \n\
Attribute \"flyby_vpn\" is not defined in %s.aoff \n\
Use format: \n\
\"flyby_prp generic fff %s.prp\" and \n\
\"flyby_vpn generic fff %s.vpn\" \n\
in %s.aoff\n\
where each file has the format: \n\
num   \n\
x y z \n\
...   \n\
...   num values \n\
...   \n\
x y z \n\
where x y z are values for PRP and the VPN \n\
\n\
Note: A spline curve will be generated from \n\
      the PRP points \n\
\0", filename, filename, filename, filename, filename );
        free( filename );
        warning( XtParent( XtParent( w ) ), str );
        free( str );
        return;
    }

    if( n_prp_points != n_vpn_points ) {
        char *str;

        str = (char *)malloc( 500 );
        sprintf( str, "# of PRP points != # of VPN points" );
        warning( XtParent( XtParent( w ) ), str );
        free( str );
        return;
    }

    if( (prp_points[0].x != proj_ref_point.x) ||
        (prp_points[0].y != proj_ref_point.y) ||
        (prp_points[0].z != proj_ref_point.z) ||
        (prp_points[n_prp_points-1].x != proj_ref_point.x) ||
        (prp_points[n_prp_points-1].y != proj_ref_point.y) ||
        (prp_points[n_prp_points-1].z != proj_ref_point.z) ) {
        char *str;

        str = (char *)malloc( 500 );
        sprintf( str, "PRP[0] != original PRP = { %.2f %.2f %.2f }",
            proj_ref_point.x, proj_ref_point.y, proj_ref_point.z );
        warning( XtParent( XtParent( w ) ), str );
        free( str );
        return;
    }

    if( (prp_points[0].x != prp_points[n_prp_points-1].x) ||
        (prp_points[0].y != prp_points[n_prp_points-1].y) ||
        (prp_points[0].z != prp_points[n_prp_points-1].z) ) {
        char *str;

        str = (char *)malloc( 500 );
        sprintf( str, "PRP[first point] != PRP[last point]" );
        warning( XtParent( XtParent( w ) ), str );
        free( str );
        return;
    }

    cursor = XCreateFontCursor( display, XC_watch );
    XDefineCursor( display, window, cursor );
    XDefineCursor( display, XtWindow( XtParent( XtParent(w) ) ), cursor );
    XFlush( display );

    spline_points = splineCurve( n_prp_points, prp_points, &n_spline_points );

    for( i=0, j=0; i<n_spline_points, j<n_prp_points; i++ ) {

        proj_ref_point.x = spline_points[i].x;
        proj_ref_point.y = spline_points[i].y;
        proj_ref_point.z = spline_points[i].z;

        view_plane_normal.x = vpn_points[j].x;
        view_plane_normal.y = vpn_points[j].y;
        view_plane_normal.z = vpn_points[j].z;

        if( prp_points[j].x == spline_points[i].x &&
            prp_points[j].y == spline_points[i].y &&
            prp_points[j].z == spline_points[i].z )
            j++;

        if( proj_ref_point.z <= front_plane )
            front_plane -= front_plane - proj_ref_point.z + 0.01;

        PEXViewMappingMatrix( view_window, &viewport, perspective,
            &proj_ref_point, view_plane, back_plane, front_plane,
            view.mapping );
        PEXViewOrientationMatrix( &view_ref_pt, &view_plane_normal,
            &view_up_vec, view.orientation );

        PEXSetTableEntries( display, rattrs.view_table, 1,1, PEXLUTView,&view );
        PEXSetTableEntries( display, rattrs.view_table, 2,1, PEXLUTView,&view );

        /* Update the display */
        redraw();
    }

    XBell( display, 0 );
    XUndefineCursor( display, window );
    XUndefineCursor( display, XtWindow( XtParent( XtParent(w) ) ) );
    XFreeCursor( display, cursor );
}


XtActionProc
CanvasAnimateCr( w, event )
Widget    w;
XEvent   *event;
{
    char *filename;

    filename = (char *)malloc( strlen( FileName ) + 1 );
    strcpy( filename, FileName );
    filename[strlen(filename)-4] = '\0';
    strcat( filename, "anim" );
    makeTrans( w, filename, NULL );
    free( filename );
}


XtActionProc
CanvasAnimateEx( w, event )
Widget    w;
XEvent   *event;
{
    char *anim_file;
    AnimScript *anim;
    extern Dialog dialog_a;
    extern char * pattern;
    extern AnimScript *readTrans();
    extern void        execTrans();

    cursor = XCreateFontCursor( display, XC_watch );
    XDefineCursor( display, window, cursor );
    XDefineCursor( display, XtWindow( XtParent( XtParent(w) ) ), cursor );
    XFlush( display );

    RetryLoad:
    anim_file = XtNewString( FileName );
    strcpy( pattern, "anim" );
    if( PopupFDialog( dialog_a, "Selection", anim_file, &anim_file ) == Okay ) {
        int l = strlen(anim_file);
        if( (anim_file[l-1] != 'm') ||
            (anim_file[l-2] != 'i') ||
            (anim_file[l-3] != 'n') ||
            (anim_file[l-4] != 'a') ) {
            warning( XtParent( XtParent( w ) ),
                "Extension \".anim\" is missing \0" );
            goto RetryLoad;
        }
        anim = readTrans( w, anim_file );
        execTrans( anim );
        free( anim_file );
    }

    XBell( display, 0 );
    XUndefineCursor( display, window );
    XUndefineCursor( display, XtWindow( XtParent( XtParent(w) ) ) );
    XFreeCursor( display, cursor );
}


XtActionProc
CanvasMorph( w, event )
Widget    w;
XEvent   *event;
{
    register short i, j, k, l;
    double t;
    extern unsigned int nvertices, nfacets;
    extern Facet  *facets;
    extern PEXColorRGB *facet_colors;
    extern Vertex *vertices, *morph_vertices;
    extern PEXStructure top_struct;
    extern int      shape_hint, ignore_edges, contour_hint, color_type;
    extern unsigned int facet_attributes, vertex_attributes;
    extern PEXMatrix init_matrix;

    Vertex *between;
    PEXStructure morph_object[11];
    PEXFacetData *facet_data;
    PEXListOfVertex *vertex_data;

    vertex_attributes = PEXGANone;

    if( morph_vertices == NULL ) {
        char *filename, *str;

        filename = (char *)malloc( strlen( FileName ) + 1 );
        strcpy( filename, FileName );
        filename[strlen(filename)-5] = '\0';

        str = (char *)malloc( 500 );
        sprintf( str, "\n\
Attribute \"morph_vertices\" is not defined in %s.aoff \n\
Use format: \n\
\"morph_vertices indexed_poly fff file.geom\" \n\
in %s.aoff\n\
where file.geom is another geometry file that has \n\
the same topology as %s.geom \n\
\0", filename, filename, filename );
        free( filename );
        warning( XtParent( XtParent( w ) ), str );
        free( str );
        return;
    }

    cursor = XCreateFontCursor( display, XC_watch );
    XDefineCursor( display, window, cursor );
    XDefineCursor( display, XtWindow( XtParent( XtParent(w) ) ), cursor );
    XFlush( display );

    between = (Vertex *) malloc ( nvertices * sizeof( Vertex ) );
    facet_data  = (PEXFacetData *)   malloc( nfacets * sizeof(PEXFacetData) );
    vertex_data = (PEXListOfVertex *)malloc( nfacets * sizeof(PEXListOfVertex));

    for( t = 0., l = 0; t <= 1.; t += 0.1, l++ ) {
        morph_object[l] = PEXCreateStructure( display );
        PEXSetLocalTransform( display, morph_object[l], PEXOCStore,
            PEXReplace, init_matrix );

        for( k = 0; k < nvertices; k++ ) {
            between[k].x = (1.- t) * vertices[k].x + t * morph_vertices[k].x;
            between[k].y = (1.- t) * vertices[k].y + t * morph_vertices[k].y;
            between[k].z = (1.- t) * vertices[k].z + t * morph_vertices[k].z;
        }

        for( j = 0; j < nfacets; j++ ) {
            PEXCoord *vtx;

            if( (facet_attributes & PEXGANormal) && 
                (facet_attributes & PEXGAColor) ) 
                facet_attributes = PEXGAColor;

            if( facet_attributes & PEXGAColor ) {
                facet_data[j].rgb.red   = facet_colors[j].red;
                facet_data[j].rgb.green = facet_colors[j].green;
                facet_data[j].rgb.blue  = facet_colors[j].blue;
            }

            vertex_data[j].count = facets[j].count;
            vtx = (PEXCoord *)malloc( vertex_data[j].count * sizeof(PEXCoord) );
            vertex_data[j].vertices.no_data = vtx;
            for(i = 0; i < vertex_data[j].count; i++, vtx++) {
                vtx->x = between[facets[j].edges[i].e - 1].x;
                vtx->y = between[facets[j].edges[i].e - 1].y;
                vtx->z = between[facets[j].edges[i].e - 1].z;
            }

            PEXFillAreaWithData( display, morph_object[l], PEXOCStore,
                shape_hint, ignore_edges,
                facet_attributes, vertex_attributes,
                color_type,
                &facet_data[j],
                vertex_data[j].count, vertex_data[j].vertices );
        }
    }

    for( l = 0; l < 11; l++ ) {
        PEXSetEditingMode( display, top_struct, PEXStructureReplace );
        PEXSetElementPtr( display, top_struct, PEXBeginning, 0 );
        PEXSetElementPtrAtLabel( display, top_struct, OBJECT_LABEL, 1 );
        PEXExecuteStructure( display, top_struct, PEXOCStore, morph_object[l] );

        /* Update the display */
        redraw();
    }

    free( facet_data );
    for(j = 0; j < nfacets; j++) {
             if( vertex_attributes & PEXGANormal & PEXGAColor )
             free( vertex_data[j].vertices.rgb_normal );
        else if( vertex_attributes & PEXGANormal )
             free( vertex_data[j].vertices.normal );
        else if( vertex_attributes & PEXGAColor )
             free( vertex_data[j].vertices.rgb );
        else
             free( vertex_data[j].vertices.no_data );
    }
    free( vertex_data );
    PEXDestroyStructures( display, 11, morph_object );

    XBell( display, 0 );
    XUndefineCursor( display, window );
    XUndefineCursor( display, XtWindow( XtParent( XtParent(w) ) ) );
    XFreeCursor( display, cursor );
}


XtActionProc
CanvasSpin( w, event )
Widget    w;
XEvent   *event;
{
    float     angle;
    float     rotation;
    PEXMatrix matrix;
    PEXMatrix temp;
    PEXMatrix temp2;
    extern PEXMatrix local_matrix;

    cursor = XCreateFontCursor( display, XC_watch );
    XDefineCursor( display, window, cursor );
    XDefineCursor( display, XtWindow( XtParent( XtParent(w) ) ), cursor );
    XFlush( display );

    PEXSetEditingMode( display, object, PEXStructureReplace );
    for( angle = 0; angle <= 50; angle += 5 ) {
        rotation = DEGRAD( angle );
        PEXRotate( PEXYAxis, rotation, matrix );
        PEXRotate( PEXXAxis, 2 * rotation, temp );
        PEXRotate( PEXZAxis, rotation, temp2 );
        PEXMatrixMult( temp, temp2, temp );
        PEXMatrixMult( matrix, temp, matrix );
        PEXMatrixMult( local_matrix, matrix, local_matrix );
        PEXSetElementPtr( display, object, PEXBeginning, 2 );
        PEXSetLocalTransform( display, object, PEXOCStore,
            PEXReplace, local_matrix );
        redraw();
    }
    for( angle = 50; angle >= 0; angle -= 5 ) {
        rotation = DEGRAD( angle );
        PEXRotate( PEXYAxis, -rotation, matrix );
        PEXRotate( PEXXAxis, 2 * -rotation, temp );
        PEXRotate( PEXZAxis, -rotation, temp2 );
        PEXMatrixMult( temp, temp2, temp );
        PEXMatrixMult( matrix, temp, matrix );
        PEXMatrixMult( local_matrix, matrix, local_matrix );
        PEXSetElementPtr( display, object, PEXBeginning, 2 );
        PEXSetLocalTransform( display, object, PEXOCStore,
            PEXReplace, local_matrix );
        redraw();
    }

    XBell( display, 0 );
    XUndefineCursor( display, window );
    XUndefineCursor( display, XtWindow( XtParent( XtParent(w) ) ) );
    XFreeCursor( display, cursor );
}


float f( val )
float val;
{
    /*  return 1. / val; */
    return val;
}


float fp( val )
float val;
{
    /*  return -1. / (val*val); */
    return 1.;
}


XtActionProc
CanvasTaper( w, event )
Widget    w;
XEvent   *event;
{
}


XtActionProc
CanvasTwist( w, event )
Widget    w;
XEvent   *event;
{
    register int i, j;
    PEXStructure     twist_object = PEXCreateStructure( display );
    PEXListOfVertex *vertex_data;
    PEXFacetData    *facet_data;
    extern unsigned  int nfacets;
    extern Facet    *facets;
    extern Vertex   *vertices;

    float theta, cos_theta, sin_theta;
    float Jac[3][3];

    facet_data  = (PEXFacetData *)   malloc( nfacets * sizeof(PEXFacetData) );
    vertex_data = (PEXListOfVertex *)malloc( nfacets * sizeof(PEXListOfVertex));

    /* ASSUMPTION: counter clockwise direction */
    for(j = 0; j < nfacets; j++) {
        PEXCoord *vtx;
        vertex_data[j].count = facets[j].count;
        vtx = (PEXCoord *)malloc( vertex_data[j].count *
            sizeof(PEXCoord) );
        vertex_data[j].vertices.no_data = vtx;
        for(i = 0; i < vertex_data[j].count; i++, vtx++) {
            vtx->x = vertices[facets[j].edges[i].e - 1].x;
            vtx->y = vertices[facets[j].edges[i].e - 1].y;
            vtx->z = vertices[facets[j].edges[i].e - 1].z;

            /* ASSUMPTION: f(x) = 1/x; f'(x) = -1/x^2 */
            theta = f(vtx->z);
            cos_theta = cos( theta );
            sin_theta = sin( theta );

            Jac[0][0] =  cos_theta;
            Jac[0][1] = -sin_theta;
            Jac[0][2] = (-vtx->x * sin_theta * fp(vtx->z)) +
                (-vtx->y * cos_theta * fp(vtx->z));
            Jac[1][0] = sin_theta;
            Jac[1][1] = cos_theta;
            Jac[1][2] =  (vtx->x * cos_theta * fp(vtx->z)) +
                (-vtx->y * sin_theta * fp(vtx->z));
            Jac[2][0] = 0.;
            Jac[2][1] = 0.;
            Jac[2][2] = 1.;

            vectorMult( vtx, Jac, vtx );
        }
        PEXFillAreaWithData( display, twist_object, PEXOCStore,
            PEXShapeConvex, PEXFalse,
            PEXGANone, PEXGANone,
            PEXColorTypeRGB,
            &facet_data[j],
            vertex_data[j].count, vertex_data[j].vertices );
    }
    PEXSetEditingMode( display, top_struct, PEXStructureReplace );
    PEXSetElementPtr( display, top_struct, PEXBeginning, 0 );
    PEXSetElementPtrAtLabel( display, top_struct, OBJECT_LABEL, 1 );
    PEXExecuteStructure( display, top_struct, PEXOCStore, twist_object );
    redraw();
}


XtActionProc
CanvasBend( w, event )
Widget    w;
XEvent   *event;
{
}


XtActionProc
CanvasStepBack( w )
Widget w;
{
    extern int recording;

    if( recording ) {
        warning( XtParent( XtParent( w ) ), "Still recording... \0" );
        return;
    }

    if( copy_head->prev != NULL ) {
        copy_head = copy_head->prev;
#ifdef PIX_FRAME
        if( (copy_head != NULL) && (copy_head->frame != (Pixmap)NULL) ) {
            XCopyArea( display, copy_head->frame, drawable, gc, 0, 0,
                  (unsigned int)xwattrs.width, (unsigned int)xwattrs.height,
                  0, 0);
#else
        if( (copy_head != NULL) && (copy_head->frame != (XImage *)NULL) ) {
            XPutImage( display, drawable, gc, copy_head->frame, 0, 0, 0, 0,
                (unsigned int)xwattrs.width, (unsigned int)xwattrs.height );
#endif		/*  */
            if( using_drawable == BUFFER ) {
                XmbufDisplayBuffers( display, 1, &drawable, 0, 0 );
                drawable = (drawable == buffers[0] ? buffers[1] : buffers[0]);
            }
            XFlush( display );
        }
    }
}


XtActionProc
CanvasPlay( w, event )
Widget    w;
XEvent   *event;
{
    extern int playing, stopping, recording;
    
    if( playing ) {
        warning( XtParent( XtParent( w ) ), "Still playing... \0" );
        return;
    }
    if( recording ) {
        warning( XtParent( XtParent( w ) ), "Still recording... \0" );
        return;
    }
    playing = 1;
    stopping = 0;
    recording = 0;

    copy_head = head;
#ifdef PIX_FRAME
    while( (copy_head != NULL) && (copy_head->frame != (Pixmap)NULL) ) {
        XCopyArea( display, copy_head->frame, drawable, gc, 0, 0,
              (unsigned int)xwattrs.width, (unsigned int)xwattrs.height,
              0, 0);
#else
    while( (copy_head != NULL) && (copy_head->frame != (XImage *)NULL) ) {
        XPutImage( display, drawable, gc, copy_head->frame, 0, 0, 0, 0,
            (unsigned int)xwattrs.width, (unsigned int)xwattrs.height );
#endif
        if( using_drawable == BUFFER ) {
            XmbufDisplayBuffers( display, 1, &drawable, 0, 0 );
            drawable = (drawable == buffers[0] ? buffers[1] : buffers[0]);
        }
        XFlush( display );
        /* usleep( 20 ); */
        copy_head = copy_head->next;
    }
    playing = 0;
}


XtActionProc
CanvasStepForw( w )
Widget w;
{
    extern int recording;

    if( recording ) {
        warning( XtParent( XtParent( w ) ), "Still recording... \0" );
        return;
    }

    if( copy_head->next != NULL ) {
        copy_head = copy_head->next;
#ifdef PIX_FRAME
        if( (copy_head != NULL) && (copy_head->frame != (Pixmap)NULL) ) {
            XCopyArea( display, copy_head->frame, drawable, gc, 0, 0,
                  (unsigned int)xwattrs.width, (unsigned int)xwattrs.height,
                  0, 0);
#else
        if( (copy_head != NULL) && (copy_head->frame != (XImage *)NULL) ) {
            XPutImage( display, drawable, gc, copy_head->frame, 0, 0, 0, 0,
                (unsigned int)xwattrs.width, (unsigned int)xwattrs.height );
#endif		/*  */
            if( using_drawable == BUFFER ) {
                XmbufDisplayBuffers( display, 1, &drawable, 0, 0 );
                drawable = (drawable == buffers[0] ? buffers[1] : buffers[0]);
            }
            XFlush( display );
        }
    }
}


XtActionProc
CanvasStop( w, event )
Widget    w;
XEvent   *event;
{
    extern int playing, stopping, recording;

    playing = 0;
    stopping = 1;
    recording = 0;
    copy_head = head;
}


XtActionProc
CanvasRecord( w, event )
Widget    w;
XEvent   *event;
{
    extern int playing, stopping, recording;

    if( playing ) {
        warning( XtParent( XtParent( w ) ), "Still playing... \0" );
        return;
    }
    if( !stopping ) {
        warning( XtParent( XtParent( w ) ), "Still playing... \0" );
        return;
    }
    if( recording ) {
        warning( XtParent( XtParent( w ) ), "Still recording... \0" );
        return;
    }
    playing = 0;
    stopping = 0;
    recording = 1;
    ppm = 0;

    if((frame = (Movie) malloc( sizeof(struct _movie) )) == NULL) {
        fprintf( stderr, "CanvasRecord: ran out of memory!\n" );
        exit(0);
    }
    frame->frame = 0;
    frame->prev = NULL;
    frame->next = NULL;
    head = frame;
}


XtActionProc
CanvasSavePPM( w, event )
Widget    w;
XEvent   *event;
{
    extern int playing, stopping, recording;
    char *filename;
    char *buf, *buf1;
    char *copy, *from, *to;
    int frameCnt = 1;
    register int i;
    int nxcolor = 256;
    XColor *xcolor = (XColor *) malloc( nxcolor * sizeof(XColor) );
    char RLut[256], GLut[256], BLut[256], ULut[256], Lut[1024];
    FILE *fp;

    if( playing ) {
        warning( XtParent( XtParent( w ) ), "Still playing... \0" );
        return( (XtActionProc)0 );
    }
    if( !stopping ) {
        warning( XtParent( XtParent( w ) ), "Still recording... \0" );
        return( (XtActionProc)0 );
    }
    if( recording ) {
        warning( XtParent( XtParent( w ) ), "Still recording... \0" );
        return( (XtActionProc)0 );
    }
    playing = 0;
    stopping = 1;
    recording = 0;
    ppm = 1;

    if( xwattrs.depth == 8 ) {
        for( i=0; i<nxcolor; i++ ) {
            xcolor[i].flags = DoRed | DoGreen | DoBlue;
            xcolor[i].pixel = i;
        }
        XQueryColors( display, cmap_info.colormap, xcolor, nxcolor );
        for( i=0; i<nxcolor; i++ ) {
            RLut[i] = xcolor[i].red;
            GLut[i] = xcolor[i].green;
            BLut[i] = xcolor[i].blue;
        }
        for( i=0; i<nxcolor; i++ ) {
            Lut[i] = i;
            ULut[i] = True;
        }
    }

    cursor = XCreateFontCursor( display, XC_watch );
    XDefineCursor( display, window, cursor );
    XDefineCursor( display, XtWindow( XtParent( XtParent(w) ) ), cursor );
    XFlush( display );

    filename = (char *) malloc(strlen(FileName)+2);
    strcpy( filename, FileName );
    filename[strlen(filename)-5] = '\0';

    buf = (char *)malloc( strlen(filename) + 10 );
    copy_head = head;
    while( (copy_head != NULL) && (copy_head->frame != 0) ) {
        sprintf( buf, "%s/%s.%d.ppm", getenv("OBJ_PATH"), filename, frameCnt );
        /* Change False to True for raw (binary) output */
        WritePPMFile( buf, True,
            copy_head->frame->data, copy_head->frame->bytes_per_line,
            xwattrs.width, xwattrs.height, xwattrs.depth,
            Lut, RLut, GLut, BLut, ULut );
        copy_head = copy_head->next;
        frameCnt++;
        fprintf( stdout, "\007" );
        fflush( stdout );
    }

    to = (char *)malloc( strlen(filename) + 10 );
    sprintf( to, "%s/%s.0.ppm", getenv("OBJ_PATH"), filename );
    from = (char *)malloc( strlen(filename) + 10 );
    sprintf( from, "%s/%s.%d.ppm", getenv("OBJ_PATH"), filename, frameCnt-1 );
    copy = (char *)malloc( strlen(from) + strlen(to) + 10 );
    sprintf( copy, "cp %s %s", from, to );
    system( copy );

    buf1 = (char *)malloc( strlen(getenv("OBJ_PATH")) + strlen(filename) + 10 );
    strcat( strcat( strcpy( buf1, getenv("OBJ_PATH") ), "/" ), filename );
    strcat( buf1, ".param" );

    if( !( fp = fopen( buf1, "w" ) ) ) {
        fprintf( stderr, "Could not open %s file for writting\n", buf1 );
        exit( 1 );
    }
    else {
        fprintf( fp, "\n\
PATTERN          I\n\
OUTPUT           %s.mpg\n\
\n\
BASE_FILE_FORMAT PPM\n\
GOP_SIZE         30\n\
SLICES_PER_FRAME 1\n\
\n\
PIXEL            FULL\n\
RANGE            10\n\
PSEARCH_ALG      EXHAUSTIVE\n\
BSEARCH_ALG      EXHAUSTIVE\n\
IQSCALE          1\n\
PQSCALE          1\n\
BQSCALE          1\n\
REFERENCE_FRAME  DECODED\n\
\n\
INPUT_CONVERT    zcat *\n\
INPUT_DIR        $OBJ_PATH\n\
INPUT\n\
%s.*.ppm.Z [0-%d]\n\
END_INPUT\n\n",
        filename, filename, frameCnt-1 );
        fclose( fp );
    }

    XBell( display, 0 );
    XUndefineCursor( display, window );
    XUndefineCursor( display, XtWindow( XtParent( XtParent(w) ) ) );
    XFreeCursor( display, cursor );

    free( buf );
    free( buf1 );
    free( filename );
    free( from );
    free( to );
    free( copy );
}


XtActionProc
CanvasSaveMPG( w, event )
Widget    w;
XEvent   *event;
{
    char *filename = (char *) malloc( strlen(FileName)+1 );
    char *buf;
    XtActionProc ap;

    if( !ppm ) {
        if( (ap = CanvasSavePPM( w, event )) == 0 )
            return;
    }
    else {
        extern int playing, stopping, recording;

        if( playing ) {
            warning( XtParent( XtParent( w ) ), "Still playing... \0" );
            return;
        }
        if( !stopping ) {
            warning( XtParent( XtParent( w ) ), "Still recording... \0" );
            return;
        }
        if( recording ) {
            warning( XtParent( XtParent( w ) ), "Still recording... \0" );
            return;
        }
        playing = 0;
        stopping = 1;
        recording = 0;
        ppm = 1;
    }

    cursor = XCreateFontCursor( display, XC_watch );
    XDefineCursor( display, window, cursor );
    XDefineCursor( display, XtWindow( XtParent( XtParent(w) ) ), cursor );
    XFlush( display );

    strcpy( filename, FileName );
    filename[strlen(filename)-5] = '\0';

    buf = (char *)malloc( strlen(filename) + 20 );
    sprintf( buf, "./utils/scalePPM %s", filename );
    system( buf );
    free( buf );

    buf = (char *)malloc( strlen(filename) + 20 );
    sprintf( buf, "./utils/makeMPG %s", filename );
    system( buf );
    free( buf );
    free( filename );

    XBell( display, 0 );
    XUndefineCursor( display, window );
    XUndefineCursor( display, XtWindow( XtParent( XtParent(w) ) ) );
    XFreeCursor( display, cursor );
}


XtActionProc
CanvasUp( w, event )
Widget    w;
XEvent   *event;
{
    extern PEXRendererAttributes rattrs;
    extern PEXCoord         view_ref_pt;
    extern PEXVector        view_up_vec;
    extern PEXVector        view_plane_normal;
    extern PEXCoord2D       view_window[2];
    extern PEXNPCSubVolume  viewport;
    extern double           view_plane, front_plane, back_plane;
    extern PEXCoord         proj_ref_point;
    extern int              perspective;
    extern PEXViewEntry     view;

    view_plane_normal.y -= 0.1;

    PEXViewMappingMatrix( view_window, &viewport, perspective,
        &proj_ref_point, view_plane, back_plane, front_plane, view.mapping );
    PEXViewOrientationMatrix( &view_ref_pt, &view_plane_normal,
        &view_up_vec, view.orientation );

    PEXSetTableEntries( display, rattrs.view_table, 1, 1, PEXLUTView, &view );
    PEXSetTableEntries( display, rattrs.view_table, 2, 1, PEXLUTView, &view );

    /* Update the display */
    redraw();
}


XtActionProc
CanvasDown( w, event )
Widget    w;
XEvent   *event;
{
    extern PEXRendererAttributes rattrs;
    extern PEXCoord         view_ref_pt;
    extern PEXVector        view_up_vec;
    extern PEXVector        view_plane_normal;
    extern PEXCoord2D       view_window[2];
    extern PEXNPCSubVolume  viewport;
    extern double           view_plane, front_plane, back_plane;
    extern PEXCoord         proj_ref_point;
    extern int              perspective;
    extern PEXViewEntry     view;

    view_plane_normal.y += 0.1;

    PEXViewMappingMatrix( view_window, &viewport, perspective,
        &proj_ref_point, view_plane, back_plane, front_plane, view.mapping );
    PEXViewOrientationMatrix( &view_ref_pt, &view_plane_normal,
        &view_up_vec, view.orientation );

    PEXSetTableEntries( display, rattrs.view_table, 1, 1, PEXLUTView, &view );
    PEXSetTableEntries( display, rattrs.view_table, 2, 1, PEXLUTView, &view );

    /* Update the display */
    redraw();
}


XtActionProc
CanvasLeft( w, event )
Widget    w;
XEvent   *event;
{
    extern PEXRendererAttributes rattrs;
    extern PEXCoord         view_ref_pt;
    extern PEXVector        view_up_vec;
    extern PEXVector        view_plane_normal;
    extern PEXCoord2D       view_window[2];
    extern PEXNPCSubVolume  viewport;
    extern double           view_plane, front_plane, back_plane;
    extern PEXCoord         proj_ref_point;
    extern int              perspective;
    extern PEXViewEntry     view;

    view_plane_normal.x -= 0.1;
    view_plane_normal.z += 0.1;

    PEXViewMappingMatrix( view_window, &viewport, perspective,
        &proj_ref_point, view_plane, back_plane, front_plane, view.mapping );
    PEXViewOrientationMatrix( &view_ref_pt, &view_plane_normal,
        &view_up_vec, view.orientation );

    PEXSetTableEntries( display, rattrs.view_table, 1, 1, PEXLUTView, &view );
    PEXSetTableEntries( display, rattrs.view_table, 2, 1, PEXLUTView, &view );

    /* Update the display */
    redraw();
}


XtActionProc
CanvasRight( w, event )
Widget    w;
XEvent   *event;
{
    extern PEXRendererAttributes rattrs;
    extern PEXCoord         view_ref_pt;
    extern PEXVector        view_up_vec;
    extern PEXVector        view_plane_normal;
    extern PEXCoord2D       view_window[2];
    extern PEXNPCSubVolume  viewport;
    extern double           view_plane, front_plane, back_plane;
    extern PEXCoord         proj_ref_point;
    extern int              perspective;
    extern PEXViewEntry     view;

    view_plane_normal.x += 0.1;
    view_plane_normal.z -= 0.1;

    PEXViewMappingMatrix( view_window, &viewport, perspective,
        &proj_ref_point, view_plane, back_plane, front_plane, view.mapping );
    PEXViewOrientationMatrix( &view_ref_pt, &view_plane_normal,
        &view_up_vec, view.orientation );

    PEXSetTableEntries( display, rattrs.view_table, 1, 1, PEXLUTView, &view );
    PEXSetTableEntries( display, rattrs.view_table, 2, 1, PEXLUTView, &view );

    /* Update the display */
    redraw();
}


/*
 * The X version simply starts up an xwd process and pipes the
 * output through a xwd-to-postscript converter. The resulting
 * postscript output is left on the given file.
 *
 * When xwd is run, it brings up a crosshair cursor for window selection.
 * The user must click this in a particular window (or the root window,
 * if all windows are desired) to dump that window.
 *
 */

XtActionProc
CanvasHardcopy( w, event )
Widget    w;
XEvent   *event;
{
    char *filename, *buf;
    extern String XwdCommand;
    extern String PrintCommand;

    cursor = XCreateFontCursor( display, XC_watch );
    XDefineCursor( display, window, cursor );

    filename = (char *) malloc(strlen(getenv("OBJ_PATH"))+strlen(FileName)+2);
    strcpy( filename, getenv("OBJ_PATH") );
    strcat( filename, "/" );
    strcat( filename, FileName );
    filename[strlen(filename)-4] = '\0';
    strcat( filename, "ps" );

    buf = (char *) malloc( strlen(filename) + 80 );
    sprintf( buf, "%s -id 0x%lx | %s > %s", XwdCommand, window,
    			  PrintCommand, filename );
    system( buf );
    free( buf );
    free( filename );

    XUndefineCursor( display, window );
    XFreeCursor( display, cursor );
}


XtActionProc
CanvasMaterialPredefQuit()
{
    extern Widget w_control_material_predef;
    XtDestroyWidget( w_control_material_predef );
}


XtActionProc
CanvasVideoQuit()
{
    extern Widget w_control_video;
    XtDestroyWidget( w_control_video );
}


XtActionProc
CanvasQuit( w, event )
Widget    w;
XEvent   *event;
{
    extern int object_defined;
    extern Widget w_dialog;
    extern Widget w_control, w_control_video;
    extern int    no_darea;
    extern int smth_changed;
    Movie curr;
    extern OFFObjDesc *Obj;

    while( head != NULL ) {
        curr = head;
        head = head->next;
#ifdef PIX_FRAME
        XFreePixmap( display, curr->frame );
#else
        free( (char *)curr->frame->data );
#endif
        free( (char *)curr );
    }

    if( smth_changed ) {
        char message[80];
        String filename;
        extern Dialog dialog_s, dialog_e;

        RetryQuit:
        filename = XtNewString( FileName );
        sprintf( message, "Save file \"%s\" before quitting?", filename );
        if( PopupDialog( dialog_s, message, filename, &filename ) == Yes ) {
            int l = strlen(filename);
            if( (filename[l-1] != 'f') ||
                (filename[l-2] != 'f') ||
                (filename[l-3] != 'o') ||
                (filename[l-4] != 'a') ) {
                warning( XtParent( XtParent( w ) ),
                    "Extension \".aoff\" is missing \0" );
                goto RetryQuit;
            }
            if( ! WriteAOFF( filename ) )
            {
                sprintf( message, "Can't write file: %s", filename );
                if( PopupDialog( dialog_e, message, NULL, NULL ) == Retry )
                    goto RetryQuit;
            }
        }
    }
    smth_changed = 0;

    if( Obj ) OFFFreeObj( Obj );

    if( !object_defined ) {
        CanvasReset();
        XFlush( display );
        XtDestroyWidget( w_control );
        if( !no_darea )
        XtDestroyWidget( w_dialog );
        else
        XDestroyWindow( display, window );
    }
    else {
        if( using_drawable == BUFFER )
            XmbufDestroyBuffers( display, window );
        else if( using_drawable == PIXMAP )
            XFreePixmap( display, pixmap );
        XDestroyWindow( display, window );
        XtDestroyApplicationContext( app_context );
        (void) exit( 0 );
    }
}


void
set_color( cbs )
char *cbs;
{
    char type[3];
    float val1, val2, val3;
    extern PEXColor colorMaterial;

    if( colorInput )
        XtRemoveInput( colorInput );
    XFlush( display );

    PEXSetEditingMode( display, top_struct, PEXStructureReplace );
    PEXSetElementPtr( display, top_struct, PEXBeginning, 0 );
    PEXSetElementPtrAtLabel( display, top_struct, MATERIAL_COLOR_LABEL, 1 );

    /*
     *  Would need to parse lines of the form:
     *  RGB %f %f %f
     *  where RGB can be either of RGB, HSV, HLS, CIE
     */

    sscanf( cbs, "%s %f %f %f\n", &type, &val1, &val2, &val3 );
    if( !strcmp( type, "RGB" ) ) {
        colorMaterial.rgb.red = val1;
        colorMaterial.rgb.green = val2;
        colorMaterial.rgb.blue = val3;
        PEXSetSurfaceColor( display, top_struct, PEXOCStore,
            PEXColorTypeRGB, &colorMaterial );
    }
    else if( !strcmp( type, "HSV" ) ) {
        colorMaterial.hsv.hue = val1;
        colorMaterial.hsv.saturation = val2;
        colorMaterial.hsv.value = val3;
        if( !colorTypeHSV ) {
            HSV hsv_values;
            RGB rgb_values;
            PEXColor temp;

            hsv_values.h = val1;
            hsv_values.s = val2;
            hsv_values.v = val3;
            rgb_values = HSVToRGB( hsv_values );
            temp.rgb.red = (float)rgb_values.r / 65535.0;
            temp.rgb.green = (float)rgb_values.g / 65535.0;
            temp.rgb.blue = (float)rgb_values.b / 65535.0;
            PEXSetSurfaceColor( display, top_struct, PEXOCStore,
                PEXColorTypeRGB, &temp );
        }
        else
            PEXSetSurfaceColor( display, top_struct, PEXOCStore,
                PEXColorTypeHSV, &colorMaterial );
    }
    else if( !strcmp( type, "HLS" ) ) {
        colorMaterial.hls.hue = val1;
        colorMaterial.hls.lightness = val2;
        colorMaterial.hls.saturation = val3;
        if( !colorTypeHLS ) {
            HLS hls_values;
            RGB rgb_values;
            PEXColor temp;

            hls_values.h = val1;
            hls_values.l = val2;
            hls_values.s = val3;
            rgb_values = HLSToRGB( hls_values );
            temp.rgb.red = (float)rgb_values.r / 65535.0;
            temp.rgb.green = (float)rgb_values.g / 65535.0;
            temp.rgb.blue = (float)rgb_values.b / 65535.0;
            PEXSetSurfaceColor( display, top_struct, PEXOCStore,
                PEXColorTypeRGB, &temp );
        }
        else
            PEXSetSurfaceColor( display, top_struct, PEXOCStore,
                PEXColorTypeHLS, &colorMaterial );
    }
    else if( !strcmp( type, "CIE" ) ) {
        colorMaterial.cie.x = val1;
        colorMaterial.cie.y = val2;
        colorMaterial.cie.z = val3;
        PEXSetSurfaceColor( display, top_struct, PEXOCStore,
            PEXColorTypeCIE, &colorMaterial );
    }

    redraw();
}


void
set_material( cbs )
char *cbs;
{
    float amb, diff, spec, exp, tran;
    char rtype[3], ctype[3];
    float rval1, rval2, rval3;
    float cval1, cval2, cval3;
    char buf[80];
    extern PEXReflectionAttributes refl_props;

    if( matInput )
        XtRemoveInput( matInput );
    XFlush( display );

    PEXSetEditingMode( display, top_struct, PEXStructureReplace );
    PEXSetElementPtr( display, top_struct, PEXBeginning, 0 );
    PEXSetElementPtrAtLabel( display, top_struct, REFL_PROPS_LABEL, 1 );

    /*
     *    Would need to parse triple pairs of lines of the form:
     *    Mat Refl Attrs Amb %f Diff %f Tran %f Spec %f Exp %f
     *    Mat Refl Color RGB %f %f %f 
     *    Mat Color RGB %f %f %f 
     *    where RGB can be either of RGB, HSV, HLS, CIE
     */

    sscanf( cbs, "Mat Refl Attrs Amb %f Diff %f Tran %f Spec %f Exp %f\nMat Refl Color %s %f %f %f\nMat Color %s %f %f %f\n",
        &amb, &diff, &tran, &spec, &exp, &rtype, &rval1, &rval2, &rval3, &ctype, &cval1, &cval2, &cval3 );
    refl_props.ambient = amb;
    refl_props.diffuse = diff;
    refl_props.transmission = tran;
    refl_props.specular = spec;
    refl_props.specular_conc = exp;
    refl_props.specular_color.type = PEXColorTypeRGB;
    if( !strcmp( rtype, "RGB" ) ) {
        SET_COLOR( rval1, rval2, rval3, refl_props.specular_color.value );
    }
    else if( !strcmp( rtype, "HSV" ) ) {
        HSV hsv_values;
        RGB rgb_values;

        hsv_values.h = rval1;
        hsv_values.s = rval2;
        hsv_values.v = rval3;
        rgb_values = HSVToRGB( hsv_values );
        SET_COLOR( (float)rgb_values.r / 65535.0, (float)rgb_values.g / 65535.0,
                   (float)rgb_values.b / 65535.0,
                   refl_props.specular_color.value );
    }
    else if( !strcmp( rtype, "HLS" ) ) {
        HLS hls_values;
        RGB rgb_values;

        hls_values.h = rval1;
        hls_values.l = rval2;
        hls_values.s = rval3;
        rgb_values = HLSToRGB( hls_values );
        SET_COLOR( (float)rgb_values.r / 65535.0, (float)rgb_values.g / 65535.0,
                   (float)rgb_values.b / 65535.0,
                   refl_props.specular_color.value );
    }
    else if( !strcmp( rtype, "CIE" ) ) {
        CIE cie_values;
        RGB rgb_values;

        cie_values.X = rval1;
        cie_values.Y = rval2;
        cie_values.Z = rval3;
        rgb_values = CIEToRGB( cie_values );
        SET_COLOR( (float)rgb_values.r, (float)rgb_values.g, (float)rgb_values.b, refl_props.specular_color.value );
    }
    PEXSetReflectionAttributes( display, top_struct, PEXOCStore, &refl_props );
    PEXSetBFReflectionAttributes( display, top_struct, PEXOCStore, &refl_props);

    sprintf( buf, "%s %f %f %f\n", ctype, cval1, cval2, cval3 );
    set_color( buf );
}


void
set_light( cbs )
char *cbs;
{
    register int i;
    extern int cur_shading, cur_refl;
    extern int numLightsON, numLightsOFF;
    extern PEXTableIndex lightsON[4], lightsOFF[4];
    extern PEXLightEntry lights[4];
    extern int light_switches[4];

    char ltype[3], rtype[3];
    float rval1, rval2, rval3;
    extern PEXRendererAttributes rattrs;
    char *temp=malloc(256);

    if( matInput )
        XtRemoveInput( matInput );
    XFlush( display );

    /*
     *  Would need to parse triple pairs of lines of the form:
     *  Lgts ON 3
     *  Dir -1.0 0.0 0.0 RGB 1.000 1.000 1.000
     *  Pos 0.7 2.0 0.7 0.0 1.0 RGB 1.000 1.000 1.000
     *  Spt 0.5 0.5 2.0 0.0 0.0 1.0 0.0 0.190 12.0 1.571 RGB 1.000 1.000 1.000
     *  Lgts OF 1
     *  Amb RGB 1.000 1.000 1.000
     *  Shad Clr
     *  Refl Spc
     *
     *  where RGB can be either of RGB, HSV, HLS, CIE
     */

    for( i=0; i<4; i++ ) lightsON[i] = 0;
    sscanf( cbs, "Lgts ON %d\n", &numLightsON );
    cbs += 10;
    for( i=0; i<numLightsON; i++ ) {
        sscanf( cbs, "%s", &ltype );
        cbs += 4;
        if( !strcmp( ltype, "Amb" ) ) {
            lightsON[i] = 1;
            light_switches[AMBIENT] = 1;
            sscanf( cbs, "%s %f %f %f\n",
                &rtype,
                &rval1, &rval2, &rval3 );
            sprintf( temp, "%.3f %.3f %.3f", rval1, rval2, rval3 );
            cbs += 5 + strlen(temp);
            if( !strcmp( rtype, "RGB" ) ) {
                SET_COLOR( rval1, rval2, rval3, lights[AMBIENT].color.value );
            }
            else if( !strcmp( rtype, "HSV" ) ) {
                HSV hsv_values;
                RGB rgb_values;

                hsv_values.h = rval1;
                hsv_values.s = rval2;
                hsv_values.v = rval3;
                rgb_values = HSVToRGB( hsv_values );
                SET_COLOR( (float)rgb_values.r / 65535.0,
                           (float)rgb_values.g / 65535.0,
                           (float)rgb_values.b / 65535.0,
                           lights[AMBIENT].color.value );
            }
            else if( !strcmp( rtype, "HLS" ) ) {
                HLS hls_values;
                RGB rgb_values;

                hls_values.h = rval1;
                hls_values.l = rval2;
                hls_values.s = rval3;
                rgb_values = HLSToRGB( hls_values );
                SET_COLOR( (float)rgb_values.r / 65535.0,
                           (float)rgb_values.g / 65535.0,
                           (float)rgb_values.b / 65535.0,
                           lights[AMBIENT].color.value );
            }
            else if( !strcmp( rtype, "CIE" ) ) {
                CIE cie_values;
                RGB rgb_values;

                cie_values.X = rval1;
                cie_values.Y = rval2;
                cie_values.Z = rval3;
                rgb_values = CIEToRGB( cie_values );
                SET_COLOR( (float)rgb_values.r,
                           (float)rgb_values.g,
                           (float)rgb_values.b,
                           lights[AMBIENT].color.value );
            }
            lights[AMBIENT].type = PEXLightAmbient;
            lights[AMBIENT].color.type = PEXColorTypeRGB;
            PEXSetTableEntries( display, rattrs.light_table, 1, 1,
                PEXLUTLight, &lights[AMBIENT] );
        }
        else if( !strcmp( ltype, "Dir" ) ) {
            lightsON[i] = 2;
            light_switches[DIRECTIONAL] = 1;
            sscanf( cbs, "%f %f %f %s %f %f %f\n",
                &lights[DIRECTIONAL].direction.x, 
                &lights[DIRECTIONAL].direction.y, 
                &lights[DIRECTIONAL].direction.z,
                &rtype,
                &rval1, &rval2, &rval3 );
            sprintf( temp, "%.3f %.3f %.3f %.3f %.3f %.3f", 
                lights[DIRECTIONAL].direction.x, 
                lights[DIRECTIONAL].direction.y, 
                lights[DIRECTIONAL].direction.z,
                rval1, rval2, rval3 );
            cbs += 5 + strlen(temp);
            if( !strcmp( rtype, "RGB" ) ) {
                SET_COLOR( rval1, rval2, rval3, lights[DIRECTIONAL].color.value );
            }
            else if( !strcmp( rtype, "HSV" ) ) {
                HSV hsv_values;
                RGB rgb_values;

                hsv_values.h = rval1;
                hsv_values.s = rval2;
                hsv_values.v = rval3;
                rgb_values = HSVToRGB( hsv_values );
                SET_COLOR( (float)rgb_values.r / 65535.0,
                           (float)rgb_values.g / 65535.0,
                           (float)rgb_values.b / 65535.0,
                           lights[DIRECTIONAL].color.value );
            }
            else if( !strcmp( rtype, "HLS" ) ) {
                HLS hls_values;
                RGB rgb_values;

                hls_values.h = rval1;
                hls_values.l = rval2;
                hls_values.s = rval3;
                rgb_values = HLSToRGB( hls_values );
                SET_COLOR( (float)rgb_values.r / 65535.0,
                           (float)rgb_values.g / 65535.0,
                           (float)rgb_values.b / 65535.0,
                           lights[DIRECTIONAL].color.value );
            }
            else if( !strcmp( rtype, "CIE" ) ) {
                CIE cie_values;
                RGB rgb_values;

                cie_values.X = rval1;
                cie_values.Y = rval2;
                cie_values.Z = rval3;
                rgb_values = CIEToRGB( cie_values );
                SET_COLOR( (float)rgb_values.r, (float)rgb_values.g,
                           (float)rgb_values.b,
                           lights[DIRECTIONAL].color.value );
            }
            lights[DIRECTIONAL].type = PEXLightWCVector;
            lights[DIRECTIONAL].color.type = PEXColorTypeRGB;
            PEXSetTableEntries( display, rattrs.light_table, 2, 1,
                PEXLUTLight, &lights[DIRECTIONAL] );
        }
        else if( !strcmp( ltype, "Pos" ) ) {
            lightsON[i] = 3;
            light_switches[POSITIONAL] = 1;
            sscanf( cbs, "%f %f %f %f %f %s %f %f %f\n",
                &lights[POSITIONAL].point.x, 
                &lights[POSITIONAL].point.y, 
                &lights[POSITIONAL].point.z,
                &lights[POSITIONAL].attenuation1,
                &lights[POSITIONAL].attenuation2,
                &rtype,
                &rval1, &rval2, &rval3 );
            sprintf( temp, "%.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f", 
                lights[POSITIONAL].point.x, 
                lights[POSITIONAL].point.y, 
                lights[POSITIONAL].point.z,
                lights[POSITIONAL].attenuation1,
                lights[POSITIONAL].attenuation2,
                rval1, rval2, rval3 );
            cbs += 5 + strlen(temp);
            if( !strcmp( rtype, "RGB" ) ) {
                SET_COLOR( rval1, rval2, rval3, lights[POSITIONAL].color.value );
            }
            else if( !strcmp( rtype, "HSV" ) ) {
                HSV hsv_values;
                RGB rgb_values;

                hsv_values.h = rval1;
                hsv_values.s = rval2;
                hsv_values.v = rval3;
                rgb_values = HSVToRGB( hsv_values );
                SET_COLOR( (float)rgb_values.r / 65535.0,
                           (float)rgb_values.g / 65535.0,
                           (float)rgb_values.b / 65535.0,
                           lights[POSITIONAL].color.value );
            }
            else if( !strcmp( rtype, "HLS" ) ) {
                HLS hls_values;
                RGB rgb_values;

                hls_values.h = rval1;
                hls_values.l = rval2;
                hls_values.s = rval3;
                rgb_values = HLSToRGB( hls_values );
                SET_COLOR( (float)rgb_values.r / 65535.0,
                           (float)rgb_values.g / 65535.0,
                           (float)rgb_values.b / 65535.0,
                           lights[POSITIONAL].color.value );
            }
            else if( !strcmp( rtype, "CIE" ) ) {
                CIE cie_values;
                RGB rgb_values;

                cie_values.X = rval1;
                cie_values.Y = rval2;
                cie_values.Z = rval3;
                rgb_values = CIEToRGB( cie_values );
                SET_COLOR( (float)rgb_values.r, (float)rgb_values.g,
                           (float)rgb_values.b,
                           lights[POSITIONAL].color.value );
            }
            lights[POSITIONAL].type = PEXLightWCPoint;
            lights[POSITIONAL].color.type = PEXColorTypeRGB;
            PEXSetTableEntries( display, rattrs.light_table, 3, 1,
                PEXLUTLight, &lights[POSITIONAL] );
        }
        else if( !strcmp( ltype, "Spt" ) ) {
            lightsON[i] = 4;
            light_switches[SPOT] = 1;
            sscanf( cbs, "%f %f %f %f %f %f %f %f %f %f %s %f %f %f\n",
                &lights[SPOT].point.x, 
                &lights[SPOT].point.y, 
                &lights[SPOT].point.z,
                &lights[SPOT].direction.x, 
                &lights[SPOT].direction.y, 
                &lights[SPOT].direction.z,
                &lights[SPOT].attenuation1,
                &lights[SPOT].attenuation2,
                &lights[SPOT].concentration,
                &lights[SPOT].spread_angle,
                &rtype,
                &rval1, &rval2, &rval3 );
            sprintf( temp, "%.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f", 
                lights[SPOT].point.x, 
                lights[SPOT].point.y, 
                lights[SPOT].point.z,
                lights[SPOT].direction.x, 
                lights[SPOT].direction.y, 
                lights[SPOT].direction.z,
                lights[SPOT].attenuation1,
                lights[SPOT].attenuation2,
                lights[SPOT].concentration,
                lights[SPOT].spread_angle,
                rval1, rval2, rval3 );
            cbs += 5 + strlen(temp);
            if( !strcmp( rtype, "RGB" ) ) {
                SET_COLOR( rval1, rval2, rval3, lights[SPOT].color.value );
            }
            else if( !strcmp( rtype, "HSV" ) ) {
                HSV hsv_values;
                RGB rgb_values;

                hsv_values.h = rval1;
                hsv_values.s = rval2;
                hsv_values.v = rval3;
                rgb_values = HSVToRGB( hsv_values );
                SET_COLOR( (float)rgb_values.r / 65535.0,
                           (float)rgb_values.g / 65535.0,
                           (float)rgb_values.b / 65535.0,
                           lights[SPOT].color.value );
            }
            else if( !strcmp( rtype, "HLS" ) ) {
                HLS hls_values;
                RGB rgb_values;

                hls_values.h = rval1;
                hls_values.l = rval2;
                hls_values.s = rval3;
                rgb_values = HLSToRGB( hls_values );
                SET_COLOR( (float)rgb_values.r / 65535.0,
                           (float)rgb_values.g / 65535.0,
                           (float)rgb_values.b / 65535.0,
                           lights[SPOT].color.value );
            }
            else if( !strcmp( rtype, "CIE" ) ) {
                CIE cie_values;
                RGB rgb_values;

                cie_values.X = rval1;
                cie_values.Y = rval2;
                cie_values.Z = rval3;
                rgb_values = CIEToRGB( cie_values );
                SET_COLOR( (float)rgb_values.r, (float)rgb_values.g,
                           (float)rgb_values.b, lights[SPOT].color.value );
            }
            lights[SPOT].type = PEXLightWCSpot;
            lights[SPOT].color.type = PEXColorTypeRGB;
            PEXSetTableEntries( display, rattrs.light_table, 4, 1,
                PEXLUTLight, &lights[SPOT] );
        }
    }

    for( i=0; i<4; i++ ) lightsOFF[i] = 0;
    sscanf( cbs, "Lgts OF %d\n", &numLightsOFF );
    cbs += 10;
    for( i=0; i<numLightsOFF; i++ ) {
        sscanf( cbs, "%s", &ltype );
        cbs += 4;
        if( !strcmp( ltype, "Amb" ) ) {
            lightsOFF[i] = 1;
            light_switches[AMBIENT] = 0;
            sscanf( cbs, "%s %f %f %f\n",
                &rtype,
                &rval1, &rval2, &rval3 );
            sprintf( temp, "%.3f %.3f %.3f", rval1, rval2, rval3 );
            cbs += 5 + strlen(temp);
            if( !strcmp( rtype, "RGB" ) ) {
                SET_COLOR( rval1, rval2, rval3, lights[AMBIENT].color.value );
            }
            else if( !strcmp( rtype, "HSV" ) ) {
                HSV hsv_values;
                RGB rgb_values;

                hsv_values.h = rval1;
                hsv_values.s = rval2;
                hsv_values.v = rval3;
                rgb_values = HSVToRGB( hsv_values );
                SET_COLOR( (float)rgb_values.r / 65535.0,
                           (float)rgb_values.g / 65535.0,
                           (float)rgb_values.b / 65535.0,
                           lights[AMBIENT].color.value );
            }
            else if( !strcmp( rtype, "HLS" ) ) {
                HLS hls_values;
                RGB rgb_values;

                hls_values.h = rval1;
                hls_values.l = rval2;
                hls_values.s = rval3;
                rgb_values = HLSToRGB( hls_values );
                SET_COLOR( (float)rgb_values.r / 65535.0,
                           (float)rgb_values.g / 65535.0,
                           (float)rgb_values.b / 65535.0,
                           lights[AMBIENT].color.value );
            }
            else if( !strcmp( rtype, "CIE" ) ) {
                CIE cie_values;
                RGB rgb_values;

                cie_values.X = rval1;
                cie_values.Y = rval2;
                cie_values.Z = rval3;
                rgb_values = CIEToRGB( cie_values );
                SET_COLOR( (float)rgb_values.r, (float)rgb_values.g,
                           (float)rgb_values.b, lights[AMBIENT].color.value );
            }
            lights[AMBIENT].type = PEXLightAmbient;
            lights[AMBIENT].color.type = PEXColorTypeRGB;
            PEXSetTableEntries( display, rattrs.light_table, 1, 1,
                PEXLUTLight, &lights[AMBIENT] );
        }
        if( !strcmp( ltype, "Dir" ) ) {
            lightsOFF[i] = 2;
            light_switches[DIRECTIONAL] = 0;
            sscanf( cbs, "%f %f %f %s %f %f %f\n",
                &lights[DIRECTIONAL].direction.x, 
                &lights[DIRECTIONAL].direction.y, 
                &lights[DIRECTIONAL].direction.z,
                &rtype,
                &rval1, &rval2, &rval3 );
            sprintf( temp, "%.3f %.3f %.3f %.3f %.3f %.3f",
                lights[DIRECTIONAL].direction.x,
                lights[DIRECTIONAL].direction.y,
                lights[DIRECTIONAL].direction.z,
                rval1, rval2, rval3 );
            cbs += 5 + strlen(temp);
            if( !strcmp( rtype, "RGB" ) ) {
                SET_COLOR( rval1, rval2, rval3, lights[DIRECTIONAL].color.value );
            }
            else if( !strcmp( rtype, "HSV" ) ) {
                HSV hsv_values;
                RGB rgb_values;

                hsv_values.h = rval1;
                hsv_values.s = rval2;
                hsv_values.v = rval3;
                rgb_values = HSVToRGB( hsv_values );
                SET_COLOR( (float)rgb_values.r / 65535.0,
                           (float)rgb_values.g / 65535.0,
                           (float)rgb_values.b / 65535.0,
                           lights[DIRECTIONAL].color.value );
            }
            else if( !strcmp( rtype, "HLS" ) ) {
                HLS hls_values;
                RGB rgb_values;

                hls_values.h = rval1;
                hls_values.l = rval2;
                hls_values.s = rval3;
                rgb_values = HLSToRGB( hls_values );
                SET_COLOR( (float)rgb_values.r / 65535.0,
                           (float)rgb_values.g / 65535.0,
                           (float)rgb_values.b / 65535.0,
                           lights[DIRECTIONAL].color.value );
            }
            else if( !strcmp( rtype, "CIE" ) ) {
                CIE cie_values;
                RGB rgb_values;

                cie_values.X = rval1;
                cie_values.Y = rval2;
                cie_values.Z = rval3;
                rgb_values = CIEToRGB( cie_values );
                SET_COLOR( (float)rgb_values.r, (float)rgb_values.g,
                           (float)rgb_values.b,
                           lights[DIRECTIONAL].color.value );
            }
            lights[DIRECTIONAL].type = PEXLightWCVector;
            lights[DIRECTIONAL].color.type = PEXColorTypeRGB;
            PEXSetTableEntries( display, rattrs.light_table, 2, 1,
                PEXLUTLight, &lights[DIRECTIONAL] );
        }
        if( !strcmp( ltype, "Pos" ) ) {
            lightsOFF[i] = 3;
            light_switches[POSITIONAL] = 0;
            sscanf( cbs, "%f %f %f %f %f %s %f %f %f\n",
                &lights[POSITIONAL].point.x, 
                &lights[POSITIONAL].point.y, 
                &lights[POSITIONAL].point.z,
                &lights[POSITIONAL].attenuation1,
                &lights[POSITIONAL].attenuation2,
                &rtype,
                &rval1, &rval2, &rval3 );
            sprintf( temp, "%.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f",
                lights[POSITIONAL].point.x,
                lights[POSITIONAL].point.y,
                lights[POSITIONAL].point.z,
                lights[POSITIONAL].attenuation1,
                lights[POSITIONAL].attenuation2,
                rval1, rval2, rval3 );
            cbs += 5 + strlen(temp);
            if( !strcmp( rtype, "RGB" ) ) {
                SET_COLOR( rval1, rval2, rval3, lights[POSITIONAL].color.value );
            }
            else if( !strcmp( rtype, "HSV" ) ) {
                HSV hsv_values;
                RGB rgb_values;

                hsv_values.h = rval1;
                hsv_values.s = rval2;
                hsv_values.v = rval3;
                rgb_values = HSVToRGB( hsv_values );
                SET_COLOR( (float)rgb_values.r / 65535.0,
                           (float)rgb_values.g / 65535.0,
                           (float)rgb_values.b / 65535.0,
                           lights[POSITIONAL].color.value );
            }
            else if( !strcmp( rtype, "HLS" ) ) {
                HLS hls_values;
                RGB rgb_values;

                hls_values.h = rval1;
                hls_values.l = rval2;
                hls_values.s = rval3;
                rgb_values = HLSToRGB( hls_values );
                SET_COLOR( (float)rgb_values.r / 65535.0,
                           (float)rgb_values.g / 65535.0,
                           (float)rgb_values.b / 65535.0,
                           lights[POSITIONAL].color.value );
            }
            else if( !strcmp( rtype, "CIE" ) ) {
                CIE cie_values;
                RGB rgb_values;

                cie_values.X = rval1;
                cie_values.Y = rval2;
                cie_values.Z = rval3;
                rgb_values = CIEToRGB( cie_values );
                SET_COLOR( (float)rgb_values.r, (float)rgb_values.g,
                           (float)rgb_values.b, lights[POSITIONAL].color.value );
            }
            lights[POSITIONAL].type = PEXLightWCPoint;
            lights[POSITIONAL].color.type = PEXColorTypeRGB;
            PEXSetTableEntries( display, rattrs.light_table, 3, 1,
                PEXLUTLight, &lights[POSITIONAL] );
        }
        if( !strcmp( ltype, "Spt" ) ) {
            lightsOFF[i] = 4;
            light_switches[SPOT] = 0;
            sscanf( cbs, "%f %f %f %f %f %f %f %f %f %f %s %f %f %f\n",
                &lights[SPOT].point.x, 
                &lights[SPOT].point.y, 
                &lights[SPOT].point.z,
                &lights[SPOT].direction.x, 
                &lights[SPOT].direction.y, 
                &lights[SPOT].direction.z,
                &lights[SPOT].attenuation1,
                &lights[SPOT].attenuation2,
                &lights[SPOT].concentration,
                &lights[SPOT].spread_angle,
                &rtype,
                &rval1, &rval2, &rval3 );
            sprintf( temp, "%.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f",
                lights[SPOT].point.x,
                lights[SPOT].point.y,
                lights[SPOT].point.z,
                lights[SPOT].direction.x,
                lights[SPOT].direction.y,
                lights[SPOT].direction.z,
                lights[SPOT].attenuation1,
                lights[SPOT].attenuation2,
                lights[SPOT].concentration,
                lights[SPOT].spread_angle,
                rval1, rval2, rval3 );
            cbs += 5 + strlen(temp);
            if( !strcmp( rtype, "RGB" ) ) {
                SET_COLOR( rval1, rval2, rval3, lights[SPOT].color.value );
            }
            else if( !strcmp( rtype, "HSV" ) ) {
                HSV hsv_values;
                RGB rgb_values;

                hsv_values.h = rval1;
                hsv_values.s = rval2;
                hsv_values.v = rval3;
                rgb_values = HSVToRGB( hsv_values );
                SET_COLOR( (float)rgb_values.r / 65535.0,
                           (float)rgb_values.g / 65535.0,
                           (float)rgb_values.b / 65535.0,
                           lights[SPOT].color.value );
            }
            else if( !strcmp( rtype, "HLS" ) ) {
                HLS hls_values;
                RGB rgb_values;

                hls_values.h = rval1;
                hls_values.l = rval2;
                hls_values.s = rval3;
                rgb_values = HLSToRGB( hls_values );
                SET_COLOR( (float)rgb_values.r / 65535.0,
                           (float)rgb_values.g / 65535.0,
                           (float)rgb_values.b / 65535.0,
                           lights[SPOT].color.value );
            }
            else if( !strcmp( rtype, "CIE" ) ) {
                CIE cie_values;
                RGB rgb_values;

                cie_values.X = rval1;
                cie_values.Y = rval2;
                cie_values.Z = rval3;
                rgb_values = CIEToRGB( cie_values );
                SET_COLOR( rgb_values.r, rgb_values.g, rgb_values.b, lights[SPOT].color.value );
            }
            lights[SPOT].type = PEXLightWCSpot;
            lights[SPOT].color.type = PEXColorTypeRGB;
            PEXSetTableEntries( display, rattrs.light_table, 4, 1,
                PEXLUTLight, &lights[SPOT] );
        }
    }

    PEXSetEditingMode( display, top_struct, PEXStructureReplace );
    PEXSetElementPtr( display, top_struct, PEXBeginning, 0 );
    PEXSetElementPtrAtLabel( display, top_struct, LIGHT_STATE_LABEL, 1 );
    PEXSetLightSourceState( display, top_struct, PEXOCStore,
        numLightsON, lightsON, numLightsOFF, lightsOFF );

    sscanf( cbs, "Shad %s", &ltype );
    cbs += 9;
    if( !strcmp( ltype, "Non" ) ) cur_shading = PEXSurfaceInterpNone;
    else if( !strcmp( ltype, "Clr" ) ) cur_shading = PEXSurfaceInterpColor;
    else if( !strcmp( ltype, "Dot" ) ) cur_shading = PEXSurfaceInterpDotProduct;
    else if( !strcmp( ltype, "Nor" ) ) cur_shading = PEXSurfaceInterpNormal;
    PEXSetEditingMode( display, top_struct, PEXStructureReplace );
    PEXSetElementPtr( display, top_struct, PEXBeginning, 0 );
    PEXSetElementPtrAtLabel( display, top_struct, SHADING_METHOD_LABEL, 1 );
    PEXSetSurfaceInterpMethod( display, top_struct, PEXOCStore, cur_shading );

    sscanf( cbs, "Refl %s", &ltype );
    if( !strcmp( ltype, "Non" ) ) cur_refl = PEXReflectionNone;
    else if( !strcmp( ltype, "Amb" ) ) cur_refl = PEXReflectionAmbient;
    else if( !strcmp( ltype, "Dif" ) ) cur_refl = PEXReflectionDiffuse;
    else if( !strcmp( ltype, "Spc" ) ) cur_refl = PEXReflectionSpecular;
    PEXSetEditingMode( display, top_struct, PEXStructureReplace );
    PEXSetElementPtr( display, top_struct, PEXBeginning, 0 );
    PEXSetElementPtrAtLabel( display, top_struct, REFL_EQN_LABEL, 1 );
    PEXSetReflectionModel( display, top_struct, PEXOCStore, cur_refl );

    redraw();
}


void get_color_reply(client_data, fid, id)
XtPointer    client_data;
int         *fid;
XtInputId   *id;
{
    char       buf[BUFSIZ];
    int        nbytes, i;

    /* Get all pending input */
    nbytes = read(*fid, buf, BUFSIZ);

    /*
     * Null terminate the string at the first newline,
     * or at the end of the bytes read.
     */
    if (nbytes && buf[0] != '\n') {
        for(i=0;i<nbytes;i++)
            if(buf[i] == '\n')
                buf[i] = '\0';
        buf[nbytes] = '\0';
        set_color( buf );
    }
}


void get_material_reply(client_data, fid, id)
XtPointer    client_data;
int         *fid;
XtInputId   *id;
{
    char       buf[BUFSIZ];
    int        nbytes, i;

    /* Get all pending input */
    nbytes = read(*fid, buf, BUFSIZ);

    /*
     * Null terminate the string at the first newline,
     * or at the end of the bytes read.
     */
    if (nbytes && buf[0] != '\n') {
        int cnt = 0;
        for(i=0;i<nbytes;i++)
            if(buf[i] == '\n')
                cnt++;
        if(buf[i] == '\n' && cnt == 3)
            buf[i] = '\0';
        buf[nbytes] = '\0';
        set_material( buf );
    }
}


void get_light_reply(client_data, fid, id)
XtPointer    client_data;
int         *fid;
XtInputId   *id;
{
    char       buf[BUFSIZ];
    int        nbytes, i;

    /* Get all pending input */
    nbytes = read(*fid, buf, BUFSIZ);

    /*
     * Null terminate the string at the first newline,
     * or at the end of the bytes read.
     */
    if (nbytes && buf[0] != '\n') {
        int cnt = 0;
        for(i=0;i<nbytes;i++)
            if(buf[i] == '\n')
                cnt++;
        if(buf[i] == '\n' && cnt == 8)
            buf[i] = '\0';
        buf[nbytes] = '\0';
        set_light( buf );
    }
}
